!!==============================================================================
! This program runs tests for the module 'PMCHW_RWG_mod'.
!
!   First a series of subroutines are defines for testing specific parts of 
!   the module. Then a sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: November 30th 2020.
!!==============================================================================


!!===========!!
! Subroutines !
!!============!=================================================================

subroutine test_solve_matrix_equation(num_tests, test_nr, PMCHW)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use is_close_mod     , only: is_close

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type), intent(in)    :: PMCHW
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(:, :), allocatable :: A
  complex(wp), dimension(:, :), allocatable :: x
  complex(wp), dimension(:, :), allocatable :: b
  complex(wp), dimension(:, :), allocatable :: Ax
  complex(wp), dimension(:, :), allocatable :: diff
  real(wp), dimension(:)   , allocatable    :: avg_abs_diff_real
  real(wp), dimension(:)   , allocatable    :: avg_abs_diff_im
  real(wp)                                  :: solver_diff_tolerance
  real(wp)                                  :: relative_diff_real
  real(wp)                                  :: relative_diff_im
  integer :: num_nonzero_b_real
  integer :: num_nonzero_b_im
  integer :: i
  integer :: j
  integer :: k
  integer :: m

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'solve_matrix_equation')

  solver_diff_tolerance = 2.E-8_wp

  allocate(A, source=PMCHW%get_PMCHW_matrix())
  allocate(b, source=PMCHW%get_q_vectors())
  !allocate(x, source=PMCHW%get_solutions())
  allocate(x(2*PMCHW%RWG_basis%num_bases, PMCHW%num_q_vectors))
  x = PMCHW%get_solutions()
  allocate(Ax, mold=b)
  allocate(diff, mold=b)
 
  Ax = matmul(A, x)
  diff = Ax - b


  allocate(avg_abs_diff_real(PMCHW%num_q_vectors))
  allocate(avg_abs_diff_im(PMCHW%num_q_vectors))
  avg_abs_diff_real = 0._wp
  avg_abs_diff_im = 0._wp
  num_nonzero_b_real = 2*PMCHW%RWG_basis%num_bases
  num_nonzero_b_im = 2*PMCHW%RWG_basis%num_bases
  do k = 1, PMCHW%num_q_vectors
     do m = 1, 2*PMCHW%RWG_basis%num_bases
        if (is_close(b(m, k)%re, 0._wp)) then
           relative_diff_real = 0._wp
           num_nonzero_b_real = num_nonzero_b_real - 1
        else
           relative_diff_real = diff(m, k)%re/b(m, k)%re
        end if
        if (is_close(b(m, k)%im, 0._wp)) then
           relative_diff_im = 0._wp
           num_nonzero_b_im = num_nonzero_b_im - 1
        else
           relative_diff_im = diff(m, k)%im/b(m, k)%im
        end if
        avg_abs_diff_real(k) = avg_abs_diff_real(k) + abs(relative_diff_real)
        avg_abs_diff_im(k) = avg_abs_diff_im(k) + abs(relative_diff_im)
     end do
     avg_abs_diff_real(k) = avg_abs_diff_real(k)/num_nonzero_b_real
     avg_abs_diff_im(k) = avg_abs_diff_im(k)/num_nonzero_b_im
     if (avg_abs_diff_real(k) > solver_diff_tolerance .or. &
          avg_abs_diff_im(k) > solver_diff_tolerance) then
        print *, 'Error: matrix solver precision too low ...'
        print *, 'q_vector:', k, ' of ', PMCHW%num_q_vectors
        print *, 'Average relative difference of real part:', avg_abs_diff_real(k)
        print *, 'Average relative difference of im part:', avg_abs_diff_im(k)
     end if
  end do

  
  call print_status_done(test_nr)
end subroutine test_solve_matrix_equation

!!------------------------------------------------------------------------------

subroutine test_calc_q_vectors(num_tests, test_nr, PMCHW_tetrahedron)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_4pnt
  use constants_mod    , only: ZERO_CMPLX

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type)                :: PMCHW_tetrahedron
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(3, 1) :: inc_E_field_ampl
  complex(wp), dimension(3, 1) :: inc_H_field_ampl
  real(wp)   , dimension(3, 1) :: inc_E_field_direction
  integer    , dimension(1)    :: inc_field_type
  integer                      :: q_vector_idx
  ! Variables used to check correctness of results -----------------------------
  complex(wp), dimension(:, :), allocatable :: q_vectors
  complex(wp), dimension(12)    :: correct_q_vectors
  complex(wp), dimension(6)     :: other_correct_q_vectors
  real(wp)   , dimension(3, 3)  :: face_coords
  real(wp)   , dimension(3)     :: p_n
  real(wp)   , dimension(2)     :: signs
  integer    , dimension(2)     :: T_n
  integer    , dimension(2)     :: free_vertices
  integer                       :: num_bases
  integer                       :: i
  integer                       :: n
  integer                       :: p

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'calc_q_vector')

  q_vector_idx = 1
  inc_E_field_ampl(:, q_vector_idx) = [ &
       cmplx(1._wp, 0._wp), &
       cmplx(1._wp, 0._wp), &
       cmplx(1._wp, 0._wp) ]
  inc_H_field_ampl(:, q_vector_idx) = [ &
       cmplx(0._wp, 0._wp), &
       cmplx(0._wp, 0._wp), &
       cmplx(1._wp, 0._wp) ]
  inc_E_field_direction(:, q_vector_idx) = [ 0._wp, 0._wp, 0._wp ]
  inc_field_type = [ 1 ]
  call PMCHW_tetrahedron%calc_q_vectors(&
       GQF_triangle_3pnt, &
       inc_E_field_ampl, &
       inc_H_field_ampl, &
       inc_E_field_direction, &
       inc_field_type)
  
  ! Check correctness of result
  num_bases = 6
  signs = [ 1._wp, -1._wp ] 
  do n = 1, PMCHW_tetrahedron%RWG_basis%num_bases
     correct_q_vectors(n) = ZERO_CMPLX
     correct_q_vectors(n + num_bases) = ZERO_CMPLX

     T_n = PMCHW_tetrahedron%RWG_basis%get_adjacent_faces(n)
     free_vertices = PMCHW_tetrahedron%RWG_basis%get_free_vertices(n)
     do p = 1, 2
        face_coords = PMCHW_tetrahedron%RWG_basis%mesh%get_face_coords(T_n(p))
        p_n = PMCHW_tetrahedron%RWG_basis%&
             mesh%get_vertex_coords(free_vertices(p))
        correct_q_vectors(n) = correct_q_vectors(n) &
             + signs(p)*( &
             1._wp/3._wp*dot_product(inc_E_field_ampl(:, q_vector_idx), ( &
             face_coords(1, :) + face_coords(2, :) + face_coords(3, :))) &
             - dot_product(inc_E_field_ampl(:, q_vector_idx), p_n) )
        correct_q_vectors(n + num_bases) = correct_q_vectors(n + num_bases) &
             + signs(p)*( &
             1._wp/3._wp*dot_product(inc_H_field_ampl(:, q_vector_idx), ( &
             face_coords(1, :) + face_coords(2, :) + face_coords(3, :))) &
             - dot_product(inc_H_field_ampl(:, q_vector_idx), p_n) )
     end do
     correct_q_vectors(n) = &
          PMCHW_tetrahedron%RWG_basis%get_basis_edge_length(n)/2._wp &
          *correct_q_vectors(n)
     correct_q_vectors(n + num_bases) = &
          PMCHW_tetrahedron%RWG_basis%get_basis_edge_length(n)/2._wp &
          *correct_q_vectors(n + num_bases)
  end do

  allocate(q_vectors, source=PMCHW_tetrahedron%get_q_vectors())
  
  if (.not. all(is_close(q_vectors(:, q_vector_idx)%re, &
       correct_q_vectors%re))) then
     print *, 'Wrong q_vector-values...'
     do i = 1, 2*PMCHW_tetrahedron%RWG_basis%num_bases
        print *, q_vectors(:, q_vector_idx), correct_q_vectors(i)
     end do
  end if

  call print_status_done(test_nr)
end subroutine test_calc_q_vectors

!!------------------------------------------------------------------------------

subroutine test_E_and_H_field_at_obs_pnt(num_tests, test_nr, PMCHW)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use constants_mod    , only: ZERO_CMPLX
  use constants_mod    , only: PI
  use constants_mod    , only: I_IMAG
  use math_funcs_mod   , only: cross_prod_3D
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_4pnt

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! NB! This test is outdated! The module is updated such that the numbers
  ! from this test do not match. I.e. this test will fail unless updated.
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type)                :: PMCHW
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(3, 1) :: inc_E_field_ampl
  complex(wp), dimension(3, 1) :: inc_H_field_ampl
  real(wp)   , dimension(3, 1) :: inc_E_field_direction
  real(wp)   , dimension(3, 1) :: inc_H_field_direction
  integer    , dimension(1)    :: inc_field_type
  integer                      :: region
  integer                      :: q_vector_idx
  logical                      :: scattered
  logical                      :: test
  real(wp)   , dimension(3)    :: observation_pnt
  complex(wp), dimension(:, :), allocatable :: E_field
  complex(wp), dimension(:, :), allocatable :: H_field
  ! Variables used to check correctness of results -----------------------------
  complex(wp), dimension(3)                 :: correct_E_field
  complex(wp), dimension(3)                 :: correct_H_field
  complex(wp), dimension(3)                 :: product
  complex(wp), dimension(3)                 :: divergence
  complex(wp), dimension(3)                 :: cross_product
  real(wp)   , dimension(3, 3)  :: face_coords
  real(wp)   , dimension(3)     :: p_n
  real(wp)   , dimension(3)     :: r1_src
  real(wp)   , dimension(3)     :: r2_src
  real(wp)   , dimension(3)     :: r3_src
  real(wp)   , dimension(3)     :: r_src
  real(wp)   , dimension(2)     :: signs
  integer    , dimension(2)     :: T_n
  integer    , dimension(2)     :: free_vertices
  complex(wp)                   :: permitivity
  complex(wp)                   :: permeability
  complex(wp)                   :: wavenumber
  real(wp)                      :: angular_frequency
  real(wp)                      :: prefactor
  integer                       :: num_bases
  integer                       :: i
  integer                       :: n
  integer                       :: p

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'E_and_H_field_at_obs_pnt')

  observation_pnt = [2._wp, 3._wp, 1.4_wp]
  region = 1
  q_vector_idx = 1
  scattered = .true.
  test = .true.
  allocate(E_field(3, PMCHW%get_num_q_vectors()))
  allocate(H_field, mold=E_field)
  call PMCHW%E_and_H_field_at_obs_pnt(&
       E_field          , &
       H_field          , &
       observation_pnt  , &
       GQF_triangle_3pnt, &
       region           , &
       scattered=scattered, &
       test=test)
  
  ! Check correctness of result ------------------------------------------------
  num_bases = PMCHW%RWG_basis%num_bases
  permitivity = PMCHW%get_permitivity(region)
  permeability = PMCHW%get_permeability(region)
  angular_frequency = PMCHW%get_angular_frequency()
  wavenumber = angular_frequency*sqrt(permeability*permitivity)

  signs = [ 1._wp, -1._wp ] 

  correct_E_field = ZERO_CMPLX
  correct_H_field = ZERO_CMPLX


  do n = 1, PMCHW%RWG_basis%num_bases
     T_n = PMCHW%RWG_basis%get_adjacent_faces(n)
     free_vertices = PMCHW%RWG_basis%get_free_vertices(n)
     prefactor = PMCHW%RWG_basis%get_basis_edge_length(n)/2._wp
     do p = 1, 2
        face_coords = PMCHW%RWG_basis%mesh%get_face_coords(T_n(p))
        r1_src = face_coords(1, :)
        r2_src = face_coords(2, :)
        r3_src = face_coords(3, :)
        p_n = PMCHW%RWG_basis%&
             mesh%get_vertex_coords(free_vertices(p))

        r_src = r1_src + r2_src + r3_src
        product = 1._wp/3._wp*r_src - p_n
        divergence = 16._wp*PI**2/3._wp*r_src - 16._wp*PI**2*p_n
        cross_product = 16._wp*PI**2/3._wp*(       &
             cross_prod_3D(observation_pnt, r_src) &
             + cross_prod_3D(r_src, p_n) )         &
             - 16._wp*PI**2*cross_prod_3D(observation_pnt, p_n)
        
        correct_E_field = correct_E_field                       &
             + prefactor*signs(p)*(                             &
             - angular_frequency*permeability/I_IMAG*(          &
             1/wavenumber**2*divergence + product )             &
             + cross_product )
        correct_H_field = correct_H_field                       &
             + prefactor*signs(p)*(                             &
             - angular_frequency*permitivity/I_IMAG*(           &
             1/wavenumber**2*divergence + product )             &
             - cross_product )
        
     end do
  end do
  if (scattered) then
     correct_E_field = -correct_E_field
     correct_H_field = -correct_H_field
  end if

  if (.not. all(is_close(E_field(:, q_vector_idx)%re,  &
       correct_E_field%re)) .or.                       &
       .not. all(is_close(E_field(:, q_vector_idx)%im, &
       correct_E_field%im)) .or.                       &
       .not. all(is_close(H_field(:, q_vector_idx)%re, &
       correct_H_field%re)) .or.                       &
       .not. all(is_close(H_field(:, q_vector_idx)%im, &
       correct_H_field%im)) ) then
     print *, 'Wrong field value at obs point...'
     do i = 1, 3
        print *, E_field(i, q_vector_idx), correct_E_field(i)
     end do
     do i = 1, 3
        print *, H_field(i, q_vector_idx), correct_H_field(i)
     end do
  end if

  call print_status_done(test_nr)
  
end subroutine test_E_and_H_field_at_obs_pnt

!!----------------------------------------------------------------------------

subroutine test_inc_E_and_H_field_at_obs_pnt(num_tests, test_nr, PMCHW, &
     permeability, &
     permitivity , &
     angular_frequency)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use PMCHW_RWG_mod    , only: eval_green_func_integrals
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use constants_mod    , only: PI
  use constants_mod    , only: I_IMAG
  use constants_mod    , only: PERMEABILITY_VACUUM
  use constants_mod    , only: PERMITIVITY_VACUUM
  use constants_mod    , only: PERMITIVITY_VACUUM
  use constants_mod    , only: ZERO_CMPLX


  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type), intent(in)    :: PMCHW
  complex(wp)          , intent(in)    :: permeability
  complex(wp)          , intent(in)    :: permitivity
  real(wp)             , intent(in)    :: angular_frequency
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(3) :: E_field
  complex(wp), dimension(3) :: H_field
  real(wp)   , dimension(3) :: observation_pnt
  integer                   :: q_vector_idx
  ! Variables used to check correctness of results -----------------------------
  complex(wp),  dimension(3) :: correct_E_field
  complex(wp),  dimension(3) :: correct_H_field
  complex(wp)             :: wavenumber

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'inc_E_and_H_field_at_obs_pnt')
  
  observation_pnt = [2._wp, 3._wp, 1.4_wp]
  wavenumber = angular_frequency*sqrt(  &
       permeability &
       *permitivity)

  q_vector_idx = 1
  call PMCHW%inc_E_and_H_field_at_obs_pnt(&
       E_field     , &
       H_field     , &
       q_vector_idx, &
       observation_pnt)
  ! Check correctness of result
  ! E_field propagating in z-direction with polarisation in x-direction
  ! H_field propagating in z-direction with polarisation in y-direction
  correct_E_field = [ &
       exp(I_IMAG*wavenumber*observation_pnt(3)), &
       ZERO_CMPLX, &
       ZERO_CMPLX ]
  correct_H_field = [ &
       ZERO_CMPLX, &
       exp(I_IMAG*wavenumber*observation_pnt(3)), &
       ZERO_CMPLX ]

  if ( .not. all(is_close(E_field%re, correct_E_field%re)) .or. &
       .not. all(is_close(H_field%re, correct_H_field%re)) ) then
     print *, 'Incorrect field values...'
     print *, 'q_idx:', q_vector_idx
     print *, E_field, correct_E_field
     print *, H_field, correct_H_field
     stop 2
  end if

  q_vector_idx = 2
  call PMCHW%inc_E_and_H_field_at_obs_pnt(&
       E_field     , &
       H_field     , &
       q_vector_idx, &
       observation_pnt)
  ! Check correctness of result
  ! E_field propagating in z-direction with polarisation in x-directino
  ! H_field propagating in z-direction with polarisation in y-directino
  correct_E_field = [ &
       ZERO_CMPLX, &
       exp(I_IMAG*wavenumber*observation_pnt(1)), &
       ZERO_CMPLX ]
  correct_H_field = [ &
       ZERO_CMPLX, &
       ZERO_CMPLX, &
       exp(I_IMAG*wavenumber*observation_pnt(1)) ]

  if ( .not. all(is_close(E_field%re, correct_E_field%re)) .or. &
       .not. all(is_close(H_field%re, correct_H_field%re)) ) then
     print *, 'Incorrect field values...'
     print *, 'q_idx:', q_vector_idx
     print *, E_field, correct_E_field
     print *, H_field, correct_H_field
     stop 2
  end if
  
  call print_status_done(test_nr)
 
end subroutine test_inc_E_and_H_field_at_obs_pnt

!!------------------------------------------------------------------------------

subroutine test_eval_green_func_integrals(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use PMCHW_RWG_mod    , only: eval_green_func_integrals
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use constants_mod    , only: PI

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  complex(wp)               :: green_func_intgr_ksi
  complex(wp)               :: green_func_intgr_eta
  complex(wp)               :: green_func_intgr_
  complex(wp)               :: grad_of_green_func_intgr_ksi
  complex(wp)               :: grad_of_green_func_intgr_eta
  complex(wp)               :: grad_of_green_func_intgr_
  real(wp), dimension(3)    :: observation_pnt
  real(wp), dimension(3, 3) :: quad_pnts
  complex(wp)               :: wavenumber
  logical                   :: obs_pnt_and_src_are_close

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'eval_green_func_equation')
  
  observation_pnt = 0._wp
  quad_pnts = 1._wp/(sqrt(3._wp)*4._wp*PI)
  ! Set wavenumber equal to 0 such that the Green's function equals 1
  wavenumber = cmplx(0._wp, 0._wp)
  obs_pnt_and_src_are_close = .false.
  
  
  call eval_green_func_integrals(&
       green_func_intgr_ksi   , &
       green_func_intgr_eta   , &
       green_func_intgr_      , &
       grad_of_green_func_intgr_ksi   , &
       grad_of_green_func_intgr_eta   , &
       grad_of_green_func_intgr_      , &
       observation_pnt   , &
       quad_pnts   , &
       GQF_triangle_3pnt, &
       wavenumber        , &
       obs_pnt_and_src_are_close, &
       faces_are_in_same_plane=.false.)

  if (is_close(green_func_intgr_ksi%re, 1._wp/3._wp) .and. &
       is_close(green_func_intgr_eta%re, 1._wp/3._wp) .and. &
       is_close(green_func_intgr_%re, 1._wp) .and. &
       is_close(grad_of_green_func_intgr_ksi%re, (4._wp*PI)**2/3._wp) .and. &
       is_close(grad_of_green_func_intgr_eta%re, (4._wp*PI)**2/3._wp) .and. &
       is_close(grad_of_green_func_intgr_%re, (4._wp*PI)**2)) then
  else
     print *, 'Errorllll'
     print *, green_func_intgr_ksi, 1._wp/3._wp
     print *, green_func_intgr_eta, 1._wp/3._wp
     print *, green_func_intgr_, 1._wp 
     print *, grad_of_green_func_intgr_ksi, (4._wp*PI)**2/3._wp
     print *, grad_of_green_func_intgr_eta, (4._wp*PI)**2/3._wp
     print *, grad_of_green_func_intgr_, (4._wp*PI)**2
     stop 2
  end if
  
  call print_status_done(test_nr)
end subroutine test_eval_green_func_integrals

!!------------------------------------------------------------------------------

subroutine test_eval_outer_integrals(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use PMCHW_RWG_mod    , only: eval_outer_integrals
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use constants_mod    , only: PI

  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! NB! This test is outdated! The module is updated such that the numbers
  ! from this test do not match. I.e. this test will fail unless updated.
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(9)    :: EFIE_integrals
  complex(wp), dimension(18)   :: MFIE_integrals
  real(wp)   , dimension(3, 3) :: outer_intgr_pnts
  real(wp)   , dimension(3, 3) :: inner_intgr_pnts
  complex(wp)                  :: wavenumber
  integer                      :: centroid_idx
  logical                      :: faces_are_close
  ! Variables for checking test results
  real(wp)   , dimension(9)    :: correct_EFIE_integrals
  real(wp)   , dimension(18)   :: correct_MFIE_integrals
  integer                      :: p, q
  real(wp)                     :: a, b, c, d, e, f, g, h, i

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'eval_green_func_equation')
  
  outer_intgr_pnts = 0._wp
  inner_intgr_pnts = 1._wp/(sqrt(3._wp)*4._wp*PI)
  ! Set wavenumber equal to 0 such that the Green's function equals 1
  wavenumber = cmplx(0._wp, 0._wp)

  faces_are_close = .false.
  p = 1
  q = 2

  call eval_outer_integrals(&
       EFIE_integrals    , &
       MFIE_integrals    , &
       outer_intgr_pnts  , &
       inner_intgr_pnts  , &
       GQF_triangle_3pnt , &
       faces_are_close   , &
       p                 , &
       q                 , &
       wavenumber)
 
  ! Check correctness ----------------------------------------------------------
  a = 1._wp/3._wp
  b = 1._wp/9._wp
  c = 16._wp*PI**2/9._wp
  d = 8._wp*PI**2/9._wp
  e = 8._wp*PI**2/3._wp
  f = 4._wp*PI**2/9._wp
  g = 16._wp*PI**2/3._wp
  h = 16._wp*PI**2
  i = 4._wp*PI**2/3._wp
  correct_EFIE_integrals = [ &
       b, &    ! 1 = ksi, ksi
       b, &    ! 2 = ksi, eta
       a, &    ! 3 = ksi, _
       b, &    ! 4 = eta, ksi
       b, &    ! 5 = eta, eta
       a, &    ! 6 = eta, _
       a, &    ! 7 = _  , ksi
       a, &    ! 8 = _  , eta
       1._wp ] ! 9 = _  , _
  correct_MFIE_integrals = [ &
       c, &    ! 1 = ksi, ksi
       c, &    ! 2 = ksi, eta
       g, &    ! 3 = ksi, _
       c, &    ! 4 = eta, ksi
       c, &    ! 5 = eta, eta
       g, &    ! 6 = eta, _
       g, &    ! 7 = _  , ksi
       g, &    ! 8 = _  , eta
       h, &    ! 9 = _  , _
       d, &    !10 = ksi ksi, ksi
       d, &    !11 = ksi ksi, eta
       e, &    !12 = ksi ksi, _
       d, &    !13 = eta eta, ksi
       d, &    !14 = eta eta, eta
       e, &    !15 = eta eta, _
       f, &    !16 = ksi eta, ksi
       f, &    !17 = ksi eta, eta
       i     ] !18 = ksi eta, _
       
  if (.not. all(is_close(EFIE_integrals%re, correct_EFIE_integrals)) .or. &
       .not. all(is_close(MFIE_integrals%re, correct_MFIE_integrals)) ) then
     print *, 'Errorsss'
     do p = 1, 9
        print *, EFIE_integrals(p)%re, correct_EFIE_integrals(p)
     end do
     do p = 1, 18
        print *, MFIE_integrals(p)%re, correct_MFIE_integrals(p)
     end do
     stop 2
  end if
  
  call print_status_done(test_nr)
end subroutine test_eval_outer_integrals

!!------------------------------------------------------------------------------

subroutine test_face_pair_integral_EFIE(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use PMCHW_RWG_mod    , only: face_pair_integral_EFIE

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(9) :: EFIE_integrals
  real(wp)   , dimension(3) :: r1_obs
  real(wp)   , dimension(3) :: r2_obs
  real(wp)   , dimension(3) :: r3_obs
  real(wp)   , dimension(3) :: r1_src
  real(wp)   , dimension(3) :: r2_src
  real(wp)   , dimension(3) :: r3_src
  real(wp)   , dimension(3) :: p_m
  real(wp)   , dimension(3) :: p_n
  complex(wp)               :: wavenumber
  complex(wp)               :: res
  ! Variables for checking test results
  real(wp)                  :: correct_res

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'face_pair_integral_EFIE')
  
  wavenumber = cmplx(2._wp, 0._wp)
  r1_obs = [0._wp, 0._wp, 0._wp]
  r2_obs = [1._wp, 0._wp, 0._wp]
  r3_obs = [1._wp, 1._wp, 0._wp]
  r1_src = [2._wp, 0._wp, 0._wp]
  r2_src = [2._wp, 0._wp, 1._wp]
  r3_src = [3._wp, 0._wp, 0._wp]
  p_m = r1_obs
  p_n = r1_src

  EFIE_integrals = [ &
       cmplx(1._wp, 0._wp), &
       cmplx(2._wp, 0._wp), &
       cmplx(3._wp, 0._wp), &
       cmplx(4._wp, 0._wp), &
       cmplx(5._wp, 0._wp), &
       cmplx(6._wp, 0._wp), &
       cmplx(7._wp, 0._wp), &
       cmplx(8._wp, 0._wp), &
       cmplx(9._wp, 0._wp)  ]
  
  res = face_pair_integral_EFIE(&
       r1_obs    , &
       r2_obs    , &
       r3_obs    , &
       r1_src    , &
       r2_src    , &
       r3_src    , &
       p_m       , &
       p_n       , &
       wavenumber, &
       EFIE_integrals)
   
  ! Check correctness ----------------------------------------------------------
  correct_res = -26.25_wp
  if (.not. is_close(res%re, correct_res)) then
     print *, 'EFIE face pair integral wrong'
     print *, res, correct_res
     !stop 2
  end if
  
  call print_status_done(test_nr)
end subroutine test_face_pair_integral_EFIE

!!------------------------------------------------------------------------------

subroutine test_face_pair_integral_MFIE(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use PMCHW_RWG_mod    , only: face_pair_integral_MFIE

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  complex(wp), dimension(18) :: MFIE_integrals
  real(wp)   , dimension(3)  :: r1_obs
  real(wp)   , dimension(3)  :: r2_obs
  real(wp)   , dimension(3)  :: r3_obs
  real(wp)   , dimension(3)  :: r1_src
  real(wp)   , dimension(3)  :: r2_src
  real(wp)   , dimension(3)  :: r3_src
  real(wp)   , dimension(3)  :: p_m
  real(wp)   , dimension(3)  :: p_n
  complex(wp)                :: wavenumber
  complex(wp)                :: res
  ! Variables for checking test results
  real(wp)                   :: correct_res

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'face_pair_integral_MFIE')
  
  wavenumber = cmplx(2._wp, 0._wp)
  r1_obs = [0._wp, 0._wp, 0._wp]
  r2_obs = [1._wp, 0._wp, 0._wp]
  r3_obs = [1._wp, 1._wp, 0._wp]
  r1_src = [2._wp, 0._wp, 0._wp]
  r2_src = [2._wp, 0._wp, 1._wp]
  r3_src = [3._wp, 0._wp, 0._wp]
  p_m = r1_obs
  p_n = r1_src

  MFIE_integrals = [ &
       cmplx(1._wp, 0._wp), &
       cmplx(2._wp, 0._wp), &
       cmplx(3._wp, 0._wp), &
       cmplx(4._wp, 0._wp), &
       cmplx(5._wp, 0._wp), &
       cmplx(6._wp, 0._wp), &
       cmplx(7._wp, 0._wp), &
       cmplx(8._wp, 0._wp), &
       cmplx(9._wp, 0._wp), &
       cmplx(10._wp, 0._wp), &
       cmplx(11._wp, 0._wp), &
       cmplx(12._wp, 0._wp), &
       cmplx(13._wp, 0._wp), &
       cmplx(14._wp, 0._wp), &
       cmplx(15._wp, 0._wp), &
       cmplx(16._wp, 0._wp), &
       cmplx(17._wp, 0._wp), &
       cmplx(18._wp, 0._wp)  ]
  
  res = face_pair_integral_MFIE(&
       r1_obs    , &
       r2_obs    , &
       r3_obs    , &
       r1_src    , &
       r2_src    , &
       r3_src    , &
       p_m       , &
       p_n       , &
       MFIE_integrals)
   
  ! Check correctness ----------------------------------------------------------
  correct_res = 2._wp
  if (.not. is_close(res%re, correct_res)) then
     print *, 'MFIE face pair integral wrong'
     print *, res, correct_res
     stop 2
  end if
  
  call print_status_done(test_nr)
end subroutine test_face_pair_integral_MFIE

!!------------------------------------------------------------------------------

subroutine test_bistatic_scattering_cross_section(&
     num_tests, &
     test_nr, &
     PMCHW, &
     BBB)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: linspace
  use test_utilities   , only: meshgrid
  use is_close_mod     , only: is_close
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use constants_mod    , only: PI
  use io_mod           , only: r8mat_write
  use io_mod           , only: get_unit
  use iso_fortran_env  , only: real64
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type), intent(in)    :: PMCHW
  logical                              :: BBB
  ! Variables for internal use -------------------------------------------------
  integer , parameter                :: q_vector_idx = 1
  integer , parameter                :: num_angles = 40
  integer , parameter                :: parallel = 1
  integer , parameter                :: orthogonal = 2
  real(wp), parameter                :: arc_radius = 30._wp
  real(wp), dimension(num_angles)    :: scattering_angles
  real(wp), dimension(num_angles, 2) :: cross_sections ! (parallel, orthogonal)
  real(wp), dimension(num_angles, 3) :: table
  real(wp)                           :: max_angle
  real(wp)                           :: uniform_space
  integer                            :: s
  character(len=255)                 :: output_filename

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'bistatic_scattering_cross_section')

  ! Create scattering angles
  ! uniformly distributed from 0-2pi
  max_angle = PI
  uniform_space = max_angle/(real(num_angles) - 1._wp)
  do s = 1, num_angles
     scattering_angles(s) = (s - 1)*uniform_space
  end do

  ! Calculate the bistatic cross-sections
  print *, ''
  do s = 1, num_angles
     cross_sections(s, 1) = PMCHW%bistatic_scattering_cross_section(&
          scattering_angles(s), &
          arc_radius         , &
          parallel          , &
          q_vector_idx       , &
          GQF_triangle_3pnt   , &
          BBB)
     cross_sections(s, 2) = PMCHW%bistatic_scattering_cross_section(&
          scattering_angles(s), &
          arc_radius         , &
          orthogonal         , &
          q_vector_idx       , &
          GQF_triangle_3pnt   , &
          BBB)
  end do

  table(:, 1) = scattering_angles
  table(:, 2) = cross_sections(:, parallel)
  table(:, 3) = cross_sections(:, orthogonal)
  if (wp == real64) then
     output_filename = 'cross_sections.txt'
     call r8mat_write(output_filename, num_angles, 3, table)
  else
     print *, 'No write routine for current precision.'
  end if
  

  call print_status_done(test_nr)
end subroutine test_bistatic_scattering_cross_section

!!------------------------------------------------------------------------------

subroutine test_results(num_tests, test_nr, PMCHW, BBB)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: linspace
  use test_utilities   , only: meshgrid
  use PMCHW_RWG_mod    , only: PMCHW_RWG_type
  use io_mod           , only: r8mat_write
  use io_mod           , only: get_unit
  use iso_fortran_env  , only: real64
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt

  implicit none
  integer              , intent(in)    :: num_tests
  integer              , intent(inout) :: test_nr
  type (PMCHW_RWG_type), intent(in)    :: PMCHW
  logical              , intent(in)    :: BBB
  ! Variables for internal use -------------------------------------------------
  integer    , parameter                       :: SPATIAL_DIM = 3
  complex(wp), dimension(:, :, :, :), allocatable :: E_field
  complex(wp), dimension(:, :, :, :), allocatable :: H_field
  complex(wp), dimension(:, :, :)   , allocatable :: inc_E_field
  complex(wp), dimension(:, :, :)   , allocatable :: inc_H_field
  real(wp)   , dimension(:, :)      , allocatable :: inc_E_field_magnitude
  real(wp)   , dimension(:, :)      , allocatable :: inc_E_field_x_component
  real(wp)   , dimension(:, :)      , allocatable :: inc_E_field_z_component
  real(wp)   , dimension(:, :)      , allocatable :: E_field_magnitude
  real(wp)   , dimension(:, :)      , allocatable :: E_field_x_component
  real(wp)   , dimension(:, :)      , allocatable :: E_field_z_component
  real(wp)   , dimension(:, :)      , allocatable :: xx
  real(wp)   , dimension(:, :)      , allocatable :: zz
  real(wp)   , dimension(:)         , allocatable :: x
  real(wp)   , dimension(:)         , allocatable :: z
  real(wp)   , dimension(SPATIAL_DIM)             :: obs_pnt
  real(wp)                                        :: y
  integer                                         :: region
  integer                                         :: q_vector_idx
  integer                                         :: num_q_vectors
  integer                                         :: m
  integer                                         :: n
  integer                                         :: i
  integer                                         :: j
  logical                                         :: scattered

  call print_status_start(num_tests, test_nr, 'PMCHW_RWG_mod', &
       'results')
  
  y = 0._wp
  x = linspace(-5._wp, 5._wp, 4)
  z = linspace(-5._wp, 5._wp, 80)
  m = size(z)
  n = size(x)
  allocate(xx(m, n))
  allocate(zz(m, n))
  call meshgrid(xx, zz, x, z)
  q_vector_idx = 1
  num_q_vectors = PMCHW%get_num_q_vectors()
  allocate(inc_E_field(m, n, SPATIAL_DIM))
  allocate(inc_H_field, mold=inc_E_field)
  allocate(E_field(m, n, SPATIAL_DIM, num_q_vectors))
  allocate(H_field, mold=E_field)
  allocate(inc_E_field_magnitude(m, n))
  allocate(inc_E_field_x_component, mold=inc_E_field_magnitude)
  allocate(inc_E_field_z_component, mold=inc_E_field_magnitude)
  allocate(E_field_magnitude, mold=inc_E_field_magnitude)
  allocate(E_field_x_component, mold=inc_E_field_magnitude)
  allocate(E_field_z_component, mold=inc_E_field_magnitude)
  
  region = 1
  scattered = .false.
  do i = 1, m
     do j = 1, n
        obs_pnt = [xx(i, j), y, zz(i, j)]
        call PMCHW%inc_E_and_H_field_at_obs_pnt(&
             inc_E_field(i, j, :), &
             inc_H_field(i, j, :), &
             q_vector_idx        , &
             obs_pnt)
        if (BBB) then
          call PMCHW%E_and_H_field_at_obs_pnt_BBB(&
             E_field(i, j, :, :) , &
             H_field(i, j, :, :) , &
             obs_pnt             , &
             GQF_triangle_3pnt   , &
             region              , &
             scattered)
       else 
          call PMCHW%E_and_H_field_at_obs_pnt(&
               E_field(i, j, :, :) , &
               H_field(i, j, :, :) , &
               obs_pnt             , &
               GQF_triangle_3pnt   , &
               region              , &
               scattered)
       end if
     end do
  end do

  inc_E_field_magnitude = norm2(abs(inc_E_field), dim=3)**2
  inc_E_field_x_component = real(inc_E_Field(:, :, 1))
  inc_E_field_z_component = real(inc_E_Field(:, :, 3))
  E_field_magnitude = norm2(abs(E_field(:, :, :, q_vector_idx)), dim=3)**2
  E_field_x_component = real(E_Field(:, :, 1, q_vector_idx))
  E_field_z_component = real(E_Field(:, :, 3, q_vector_idx))
  if (wp == real64) then
     call r8mat_write('inc_E_field_magnitude.txt', m, n, inc_E_field_magnitude)
     call r8mat_write('inc_E_field_x_component.txt', m, n, &
          inc_E_field_x_component)
     call r8mat_write('inc_E_field_z_component.txt', m, n, &
          inc_E_field_z_component)
     call r8mat_write('E_field_magnitude.txt', m, n, E_field_magnitude)
     call r8mat_write('E_field_x_component.txt', m, n, E_field_x_component)
     call r8mat_write('E_field_z_component.txt', m, n, E_field_z_component)
  else
     print *, 'No write routine for current precision.'
  end if

  call print_status_done(test_nr)
end subroutine test_results

!!------------------------------------------------------------------------------

!!============!!
! Main program !
!!=============!================================================================
program main
  use working_precision      , only: wp
  use mesh_mod               , only: mesh_type
  use RWG_basis_mod          , only: RWG_basis_type
  use PMCHW_RWG_mod          , only: PMCHW_RWG_type
  ! For scattering problem example
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_4pnt
  use constants_mod          , only: PERMEABILITY_VACUUM
  use constants_mod          , only: PERMITIVITY_VACUUM
  use constants_mod          , only: PI
  use constants_mod          , only: ZERO_CMPLX
  use is_close_mod           , only: is_close
  use test_utilities         , only: print_error_message
  use test_utilities         , only: print_status_start
  use test_utilities         , only: print_status_done
  use ieee_arithmetic        , only: ieee_is_finite
  
  implicit none
  integer, parameter        :: num_tests = 20
  integer                   :: test_nr = 1
  character(:), allocatable :: module_name 
  logical, parameter        :: BBB = .false.
  ! Variables for generating mesh and basis ------------------------------------
  type (mesh_type)                       :: mesh1
  type (mesh_type)                       :: mesh2
  type (mesh_type)                       :: mesh3
  type (mesh_type)                       :: mesh4
  type (RWG_basis_type)                  :: tetrahedron
  type (RWG_basis_type)                  :: closed_cube
  type (RWG_basis_type)                  :: unit_sphere_3
  type (RWG_basis_type)                  :: unit_sphere_1
  character(len=:), allocatable          :: filename
  character(len=:), allocatable          :: file_format
  logical                                :: closed_surface
  integer                                :: edge_order
  integer                                :: spatial_dim
  integer, parameter                     :: num_face_vertices = 3
  integer                                :: num_handles
  integer                                :: num_apertures
  integer                                :: num_boundary_edges
  real(wp)                               :: scale_factor
  ! Variables for generating scattering problem example ------------------------
  type (PMCHW_RWG_type)             :: PMCHW
  type (PMCHW_RWG_type)             :: PMCHW_tetrahedron
  type (PMCHW_RWG_type)             :: PMCHW_closed_cube
  complex(wp), dimension(2)         :: permeabilities
  complex(wp), dimension(2)         :: permitivities
  real(wp)                          :: angular_frequency
  real(wp)                          :: wavelength
  complex(wp), dimension(3, 2)      :: E_inc_ampl
  complex(wp), dimension(3, 2)      :: H_inc_ampl
  real(wp)   , dimension(3, 2)      :: E_inc_direction
  real(wp)   , dimension(3, 2)      :: H_inc_direction
  integer    , dimension(2)         :: inc_field_type

  module_name = 'PMCHW_RWG_mod'
 
  !!----------------!!
  ! Tetrahedron      !
  !------------------!----------------------------------------------------------
  filename = 'meshes/tetrahedron.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0
  scale_factor = 10.e-7_wp

  call mesh1%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  ! Scale size of mesh to order of tens of nanometers
  !call mesh1%scale_nodes(scale_factor)
  call tetrahedron%initialise(mesh1)

  !!----------------!!
  ! Closed cube      !
  !------------------!----------------------------------------------------------
  filename = 'meshes/cube_closed.msh'
  call mesh4%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call closed_cube%initialise(mesh4)

  !!-------------------------!!
  ! Unit sphere (DOF = 10707) !
  !---------------------------!-------------------------------------------------
  filename = 'meshes/gmsh_sphere_1ord_3.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0

  call mesh2%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call unit_sphere_3%initialise(mesh2)

  !!-------------------------!!
  ! Unit sphere (DOF = 441)   !
  !---------------------------!-------------------------------------------------
  filename = 'meshes/gmsh_sphere_1ord.msh'

  call mesh3%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call unit_sphere_1%initialise(mesh3)

  !!------------------!!
  ! Scattering example !
  !--------------------!--------------------------------------------------------
  permeabilities(1) = cmplx(1._wp, 0._wp)*PERMEABILITY_VACUUM
  permitivities(1) = cmplx(1._wp, 0._wp)*PERMITIVITY_VACUUM
  ! Region 2, inner region. Assuming non-magnetic material, such that
  ! the relative permeability equals 1. In this case we have a complex
  ! permitivity, depending on the angular frequency of the EM-wave. 
  ! Below: High permitivity scatterer (refractive index = 4)
  permeabilities(2) = cmplx(1._wp, 0._wp)*PERMEABILITY_VACUUM
  permitivities(2) = cmplx(16._wp, 0._wp)*PERMITIVITY_VACUUM
  wavelength = 2.e-0_wp ! Wavelength in meter in vacuum
  ! With radius of the sphere, a, equal to 1 m this makes the relationship
  ! a/wavelength = 1/2
  angular_frequency = 2._wp*PI*3.e8_wp/wavelength
  
  ! The complex value below is gathered from main article for gold and
  ! wavelength = 662 nm.
  ! Refractive index = 0.138 + 2.636i
  !permeabilities(2) = cmplx(1._wp, 0._wp)*PERMEABILITY_VACUUM
  !permitivities(2) = cmplx(-13.86_wp, 1.028_wp)*PERMITIVITY_VACUUM
  !angular_frequency = 2847364942830628.5_wp ! Wavelength = 662 nm in vacuum

    ! Incoming plane waves traveling in the positive z-direction
  inc_field_type(1) = 1
  E_inc_direction(:, 1) = [ 0._wp, 0._wp, 1._wp ]
  ! Incoming plane waves traveling in the positive z-direction
  inc_field_type(2) = 1
  E_inc_direction(:, 2) = [ 1._wp, 0._wp, 0._wp ]

  ! Polarization and intensity.
  ! Illumination intensity is equal to unity.
  ! E-field is x-polarized (H-field y-polarized)
  E_inc_ampl(:, 1) = [      &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp)         , &
       cmplx(0._wp, 0._wp) ]
  H_inc_ampl(:, 1) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp) ]
  ! E-field is y-polarized (H-field z-polarized)
  E_inc_ampl(:, 2) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp) ]
  H_inc_ampl(:, 2) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp) ]

  
  call PMCHW%initialise(&
       unit_sphere_1 , &
       permeabilities, &
       permitivities , &
       angular_frequency)
  call PMCHW_tetrahedron%initialise(&
       tetrahedron   , &
       permeabilities, &
       permitivities , &
       angular_frequency)
  call PMCHW_closed_cube%initialise(&
       closed_cube   , &
       permeabilities, &
       permitivities , &
       angular_frequency)
  if (BBB) then
!!$     call PMCHW%calc_q_vectors_BBB(&
!!$          GQF_triangle_3pnt, &
!!$          E_inc_ampl       , &
!!$          H_inc_ampl       , &
!!$          E_inc_direction  , &
!!$          inc_field_type)
!!$     call PMCHW_tetrahedron%calc_q_vectors_BBB(&
!!$          GQF_triangle_3pnt, &
!!$          E_inc_ampl       , &
!!$          H_inc_ampl       , &
!!$          E_inc_direction  , &
!!$          inc_field_type)
     call PMCHW_closed_cube%calc_q_vectors_BBB(&
          GQF_triangle_3pnt, &
          E_inc_ampl       , &
          H_inc_ampl       , &
          E_inc_direction  , &
          inc_field_type)
  else
!!$    call PMCHW%calc_q_vectors(&
!!$          GQF_triangle_3pnt, &
!!$          E_inc_ampl       , &
!!$          H_inc_ampl       , &
!!$          E_inc_direction  , &
!!$          inc_field_type)
!!$     call PMCHW_tetrahedron%calc_q_vectors(&
!!$          GQF_triangle_3pnt, &
!!$          E_inc_ampl       , &
!!$          H_inc_ampl       , &
!!$          E_inc_direction  , &
!!$          inc_field_type)
     call PMCHW_closed_cube%calc_q_vectors(&
          GQF_triangle_3pnt, &
          E_inc_ampl       , &
          H_inc_ampl       , &
          E_inc_direction  , &
          inc_field_type)
  end if
!!$  print *, 'sphere'
!!$  call PMCHW%calc_PMCHW_matrix(GQF_triangle_3pnt, .true.)
!!$  call PMCHW%solve_matrix_equation()
!!$  call PMCHW_tetrahedron%calc_PMCHW_matrix(GQF_triangle_3pnt, BBB)
!!$  call PMCHW_tetrahedron%solve_matrix_equation()
  call PMCHW_closed_cube%calc_PMCHW_matrix(GQF_triangle_3pnt, BBB)
  call PMCHW_closed_cube%solve_matrix_equation()



  ! Call test functions
  !call test_solve_matrix_equation(num_tests, test_nr, PMCHW_closed_cube)
  !call test_calc_q_vectors(num_tests, test_nr, PMCHW_closed_cube)
  !call test_E_and_H_field_at_obs_pnt(num_tests, test_nr, PMCHW_closed_cube)
  call test_inc_E_and_H_field_at_obs_pnt(num_tests, test_nr, &
       PMCHW_closed_cube, &
       permeabilities(1), &
       permitivities(1) , &
       angular_frequency)
  call test_eval_green_func_integrals(num_tests, test_nr)
  call test_eval_outer_integrals(num_tests, test_nr)
  !call test_face_pair_integral_EFIE(num_tests, test_nr)
  call test_face_pair_integral_MFIE(num_tests, test_nr)
  call test_bistatic_scattering_cross_section(num_tests, test_nr, &
       PMCHW_closed_cube, BBB)
  !call test_results(num_tests, test_nr, PMCHW_closed_cube, BBB)
  call test_surface_intgr_solution(num_tests, test_nr)
  !call test_line_intgr_solution(num_tests, test_nr)
  call test_map_GLQF_pnt_to_triangle_edge(num_tests, test_nr)
  call test_calc_edge_unit_normals(num_tests, test_nr)
  !call test_inner_intgr_of_subtr_terms(num_tests, test_nr)
  !call test_green_func_subtracted(num_tests, test_nr)
  !call test_eval_subtracted_terms(num_tests, test_nr, PMCHW)
  !call test_dbl_singularity_intgr(num_tests, test_nr, PMCHW)
  !call test_get_edge_lengths(num_tests, test_nr, PMCHW_closed_cube)
  !call test_face_centroid(num_tests, test_nr, PMCHW_closed_cube)
  !call test_are_obs_pnt_and_src_close(num_tests, test_nr, PMCHW_closed_cube)


  !=======!=========================!==========================================!
contains  ! /\/\/\/\/\/\/\/\/\/\/\/\!/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\!
  !=======!=========================!==========================================!

  subroutine test_surface_intgr_solution(num_tests, test_nr)
    use PMCHW_RWG_mod, only: surface_intgr_solution
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    integer , parameter         :: NUM_FACE_VERTICES = 3
    real(wp), dimension(NUM_FACE_VERTICES) :: t1
    real(wp), dimension(NUM_FACE_VERTICES) :: t2
    real(wp), dimension(NUM_FACE_VERTICES) :: t3
    real(wp), dimension(NUM_FACE_VERTICES) :: R_0
    real(wp), dimension(NUM_FACE_VERTICES) :: R_plus1
    real(wp), dimension(NUM_FACE_VERTICES) :: R_plus2
    real(wp), dimension(NUM_FACE_VERTICES) :: R_plus3
    real(wp), dimension(NUM_FACE_VERTICES) :: R_minus1
    real(wp), dimension(NUM_FACE_VERTICES) :: R_minus2
    real(wp), dimension(NUM_FACE_VERTICES) :: s_plus1
    real(wp), dimension(NUM_FACE_VERTICES) :: s_plus2
    real(wp), dimension(NUM_FACE_VERTICES) :: s_plus3
    real(wp), dimension(NUM_FACE_VERTICES) :: s_minus1
    real(wp), dimension(NUM_FACE_VERTICES) :: s_minus2
    real(wp), dimension(NUM_FACE_VERTICES) :: s_minus3
    real(wp)                    :: h
    real(wp)                    :: w
    real(wp)                    :: x1
    real(wp)                    :: x2
    real(wp)                    :: x3
    real(wp)                    :: x4
    real(wp)                    :: x5
    real(wp)                    :: x6
    
    test_name = 'surface_intgr_solution'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    h = 1._wp
    w = 2._wp
    R_0 = [1._wp, 1._wp, 1._wp]
    t1 = [0._wp, 0._wp, 0._wp]
    t2 = [1._wp, 1._wp, 1._wp]
    t3 = [1._wp, 2._wp, 3._wp]
    R_plus1 = [0._wp, 0._wp, 0._wp]
    R_plus2 = [1._wp, 1._wp, 1._wp]
    R_plus3 = [2._wp, 2._wp, 2._wp]
    R_minus1 = [0._wp, 0._wp, 0._wp]
    R_minus2= [1._wp, 1._wp, 1._wp]
    s_plus1 = [0._wp, 0._wp, 0._wp]
    s_plus2 = [1._wp, 1._wp, 1._wp]
    s_plus3 = [2._wp, 2._wp, 2._wp]
    s_minus1 = [0._wp, 0._wp, 0._wp]
    s_minus2 = [1._wp, 1._wp, 1._wp]
    s_minus3 = [2._wp, 2._wp, 2._wp]

    !                           h t   R_0  R^+      R^-       s^+      s^-
    x1 = surface_intgr_solution(h,t1, R_0, R_plus1, R_minus1, s_plus1, s_minus1)
    x2 = surface_intgr_solution(h,t2, R_0, R_plus2, R_minus2, s_plus3, s_minus2)
    x3 = surface_intgr_solution(w,t2, R_0, R_plus2, R_minus2, s_plus2, s_minus3)
    x4 = surface_intgr_solution(h,t2, R_0, R_plus3, R_minus2, s_plus2, s_minus2)
    x5 = surface_intgr_solution(h,t2, R_0, R_plus2, R_minus2, s_plus2, s_minus2)
    x6 = surface_intgr_solution(h,t3, R_0, R_plus2, R_minus2, s_plus3, s_minus2)
    if ( .not. is_close(x1, 0._wp) .or. &
         .not. is_close(x2, 3._wp*0.3217505543966422_wp) .or. &
         .not. is_close(x3, (-0.399378073726388_wp)) .or. &
         .not. is_close(x4, 3._wp*(-0.1418970546041639_wp)) .or. &
         .not. is_close(x5, 0._wp) .or. &
         .not. is_close(x6, 0.9097531579442096_wp) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    call print_status_done(test_nr)
  end subroutine test_surface_intgr_solution
  
  !!----------------------------------------------------------------------------

  subroutine test_line_intgr_solution(num_tests, test_nr)
    use PMCHW_RWG_mod, only: line_intgr_solution
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    real(wp) :: res1
    real(wp) :: res2
    real(wp) :: res3

    test_name = 'line_intgr_solution'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)
    !                          R^+    R^-    s^+    s^-
    res1 = line_intgr_solution(0._wp, 1._wp, 0._wp, 1._wp)
    res2 = line_intgr_solution(0._wp, 5._wp, 1._wp, 5._wp)
    res3 = line_intgr_solution(1._wp, 1._wp, 1._wp, -2._wp)

    if (ieee_is_finite(res1) .or. &
         .not. is_close(res2, -2.3025850929940455_wp) .or. &
         .not. isnan(res3) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    call print_status_done(test_nr)
  end subroutine test_line_intgr_solution
  
  !!----------------------------------------------------------------------------
  
  subroutine test_map_GLQF_pnt_to_triangle_edge(num_tests, test_nr)
    use PMCHW_RWG_mod          , only: map_GLQF_pnt_to_triangle_edge
    use gauss_quad_formulas_mod, only: GQF_legendre_3pnt
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    real(wp), dimension(SPATIAL_DIM) :: vertex1
    real(wp), dimension(SPATIAL_DIM) :: vertex2
    real(wp), dimension(SPATIAL_DIM) :: vertex3
    real(wp), dimension(SPATIAL_DIM) :: edge1
    real(wp), dimension(SPATIAL_DIM) :: edge2
    real(wp), dimension(SPATIAL_DIM) :: edge3
    real(wp), dimension(SPATIAL_DIM) :: res1
    real(wp), dimension(SPATIAL_DIM) :: res2
    real(wp), dimension(SPATIAL_DIM) :: res3
    integer                          :: error_nr

    test_name = 'map_GLQF_pnt_to_triangle_edge'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    ! Triangle in xy-plane
    vertex1 = [ -1._wp, 0._wp, 0._wp ]
    vertex2 = [ 1._wp, 0._wp, 0._wp ]
    vertex3 = [ 1._wp, 1._wp, 0._wp ]
    edge1 = vertex3 - vertex2
    edge2 = vertex1 - vertex3
    edge3 = vertex2 - vertex1
    
    res1 = map_GLQF_pnt_to_triangle_edge(&
         GQF_legendre_3pnt, &
         edge1            , &
         edge2            , &
         edge3            , &
         vertex1          , &
         vertex2          , &
         vertex3          , &
         1                , &
         3)
    res2 = map_GLQF_pnt_to_triangle_edge(&
         GQF_legendre_3pnt, &
         edge1            , &
         edge2            , &
         edge3            , &
         vertex1          , &
         vertex2          , &
         vertex3          , &
         2                , &
         2)
    res3 = map_GLQF_pnt_to_triangle_edge(&
         GQF_legendre_3pnt, &
         edge1            , &
         edge2            , &
         edge3            , &
         vertex1          , &
         vertex2          , &
         vertex3          , &
         3                , &
         1)

    if (.not. all(is_close(res1, [GQF_legendre_3pnt(1, 2), 0._wp, 0._wp]))) then
       error_nr = 1
    else if (.not. all(is_close(res2, [ 0._wp, 0.5_wp, 0._wp ])) ) then
       error_nr = 2
    else if (.not. all(is_close(res3, [ 1._wp, 0.8872983346_wp, 0._wp ])) ) then
       error_nr = 3
    else
       error_nr = 0
    end if
    if (error_nr /= 0) then
       write (msg, "(A5, I2, A8)") 'Test ', error_nr, ' failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    call print_status_done(test_nr)
  end subroutine test_map_GLQF_pnt_to_triangle_edge
  
  !!----------------------------------------------------------------------------
  
  subroutine test_calc_edge_unit_normals(num_tests, test_nr)
    use PMCHW_RWG_mod, only: calc_edge_unit_normals
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable        :: test_name 
    character(len=255)               :: msg
    integer , parameter                                 :: NUM_FACE_VERTICES = 3
    integer , parameter                                 :: SPATIAL_DIM = 3
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: res1
    real(wp), dimension(NUM_FACE_VERTICES)              :: edge_lengths
    real(wp), dimension(SPATIAL_DIM)                    :: face_unit_normal
    real(wp), dimension(SPATIAL_DIM)                    :: vertex1
    real(wp), dimension(SPATIAL_DIM)                    :: vertex2
    real(wp), dimension(SPATIAL_DIM)                    :: vertex3
    
    test_name = 'calc_edge_unit_normals'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    vertex1 = [ 0._wp, 0._wp, 0._wp ]
    vertex2 = [ 1._wp, 0._wp, 0._wp ]
    vertex3 = [ 1._wp, 1._wp, 0._wp ]
    edge_lengths = [ 1._wp, sqrt(2._wp), 1._wp ]
    face_unit_normal = [ 0._wp, 0._wp, 1._wp ]

    res1 = calc_edge_unit_normals(&
         edge_lengths    , &
         face_unit_normal, &
         vertex1         , &
         vertex2         , &
         vertex3)
         
    if ( .not. all(is_close(res1(1, :), [ 1._wp, 0._wp, 0._wp ])) .or. &
         .not. all(is_close(res1(2, :), &
         [ -1._wp/sqrt(2._wp), 1._wp/sqrt(2._wp), 0._wp ])) .or. &
         .not. all(is_close(res1(3, :), [ 0._wp, -1._wp, 0._wp ])) ) then
       msg = 'Test failed.'
       call print_error_message(module_name, test_name, msg)
    end if
    call print_status_done(test_nr)
  end subroutine test_calc_edge_unit_normals
  
  !!----------------------------------------------------------------------------
  
  subroutine test_inner_intgr_of_subtr_terms(num_tests, test_nr)
    use PMCHW_RWG_mod, only: inner_intgr_of_subtr_terms
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable        :: test_name 
    character(len=255)               :: msg
    integer , parameter              :: NUM_FACE_VERTICES = 3
    integer , parameter              :: SPATIAL_DIM = 3
    real(wp)                         :: X1_minus_1
    real(wp)                         :: X1_plus_1
    real(wp), dimension(SPATIAL_DIM) :: hnX1_minus_3
    real(wp), dimension(SPATIAL_DIM) :: X2_minus_1
    real(wp), dimension(SPATIAL_DIM) :: X2_plus_1
    real(wp), dimension(SPATIAL_DIM) :: X3_minus_1
    real(wp), dimension(SPATIAL_DIM) :: X3_plus_1
    real(wp), dimension(SPATIAL_DIM) :: X4_minus_1
    real(wp), dimension(SPATIAL_DIM) :: X4_plus_1
    real(wp)                         :: correct_X1_minus
    real(wp)                         :: correct_X1_plus
    real(wp), dimension(SPATIAL_DIM) :: correct_X2_minus
    real(wp), dimension(SPATIAL_DIM) :: correct_X2_plus
    real(wp), dimension(SPATIAL_DIM) :: correct_X4_minus
    real(wp), dimension(SPATIAL_DIM) :: correct_X4_plus
    real(wp), dimension(SPATIAL_DIM) :: observation_pnt
    real(wp), dimension(SPATIAL_DIM) :: r1_src
    real(wp), dimension(SPATIAL_DIM) :: r2_src
    real(wp), dimension(SPATIAL_DIM) :: r3_src
    real(wp), dimension(SPATIAL_DIM) :: p_vec
    real(wp)                         :: face_area
    real(wp), dimension(SPATIAL_DIM) :: face_unit_normal
    real(wp), dimension(SPATIAL_DIM) :: edge_lengths
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: edge_unit_normals
    
    test_name = 'inner_intgr_of_subtr_terms'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    observation_pnt = [ -1._wp, 0._wp, 0._wp ]
    r1_src = [ -1._wp, 0._wp, 0._wp ]
    r2_src = [ 1._wp, 0._wp, 0._wp ]
    r3_src = [ 1._wp, 1._wp, 1._wp ]
    p_vec = r3_src
    face_area = sqrt(2._wp)
    face_unit_normal = [ 0._wp, -1._wp, 1._wp ]/sqrt(2._wp)
    edge_lengths = [ sqrt(2._wp), sqrt(6._wp), 2._wp ]
    edge_unit_normals(1, :) = [ 1._wp, 0._wp, 0._wp ]
    edge_unit_normals(2, :) = [ -1._wp, 1._wp, 1._wp ]/sqrt(3._wp)
    edge_unit_normals(3, :) = [ 0._wp, -1._wp, -1._wp ]/sqrt(2._wp)
    call inner_intgr_of_subtr_terms(&
         hnX1_minus_3    , &
         X1_minus_1      , &
         X1_plus_1       , &
         X2_minus_1      , &
         X2_plus_1       , &
         X3_minus_1      , &
         X3_plus_1       , &
         X4_minus_1      , &
         X4_plus_1       , &
         .false.         , &
         observation_pnt , &
         r1_src          , &
         r2_src          , &
         r3_src          , &
         p_vec           , &
         face_area       , &
         face_unit_normal, &
         edge_lengths    , &
         edge_unit_normals)

    correct_X1_minus = 1.3169578969248166_wp
    correct_X1_plus = -2.0326724696624625_wp
    correct_X2_minus = [          &
         -1.3169578969248170_wp , &
         -0.99912065172903386_wp, &
         -0.99912065172903386_wp ]
    correct_X2_plus = [         &
         7.1143536438186192_wp, &
         2.8219142356492761_wp, &
         2.8219142356492761_wp ]
    correct_X4_minus = [          &
         0._wp                  , &
         -0.65847894846240829_wp, &
         0.65847894846240829_wp ]
    correct_X4_plus = [          &
         0._wp                 , &
         -0.6812830653325070_wp, &
         0.6812830653325070_wp ]

    if ( .not. is_close(X1_minus_1, correct_X1_minus) .or. &
         .not. is_close(X1_plus_1, correct_X1_plus) .or. &
         .not. all(is_close(X2_minus_1, correct_X2_minus)) .or. &
         .not. all(is_close(X2_plus_1, correct_X2_plus)) .or. &
         .not. all(is_close(X4_minus_1, correct_X4_minus)) .or. &
         .not. all(is_close(X4_plus_1, correct_X4_plus)) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
         
    call print_status_done(test_nr)
  end subroutine test_inner_intgr_of_subtr_terms
  
  !!----------------------------------------------------------------------------

  subroutine test_green_func_subtracted(num_tests, test_nr)
    use PMCHW_RWG_mod, only: green_func_smoothened
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    complex(wp)                 :: G_subtracted
    complex(wp)                 :: grad_of_G_subtracted
    complex(wp)                 :: wavenumber
    complex(wp)                 :: kR
    real(wp)                    :: R
    logical                     :: faces_are_on_same_plane
    
    test_name = 'green_func_subtracted'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    ! Test 1
    wavenumber = cmplx(2._wp, 0.5_wp)
    R = 3._wp
    kR = wavenumber*R
    faces_are_on_same_plane = .false.
    call green_func_subtracted(&
         G_subtracted, &
         grad_of_G_subtracted, &
         wavenumber          , &
         kR                  , &
         R                   , &
         faces_are_on_same_plane)
    if ( .not. is_close(G_subtracted%re, 0.4267804243453136_wp) .or. &
         .not. is_close(G_subtracted%im, 0.2370786349647145_wp) .or. &
         .not. is_close(grad_of_G_subtracted%re, -0.05220715027370645_wp) .or. &
         .not. is_close(grad_of_G_subtracted%im, -0.03077385425652303_wp) ) then
       msg = 'Test 1 failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    
    ! Test 2
    wavenumber = cmplx(2._wp, 0._wp)
    kR = wavenumber*R
    call green_func_subtracted(&
         G_subtracted, &
         grad_of_G_subtracted, &
         wavenumber          , &
         kR                  , &
         R                   , &
         faces_are_on_same_plane)
    if ( .not. is_close(G_subtracted%re, 0.47640831331543143_wp) .or. &
         .not. is_close(G_subtracted%im, -0.007411726285807272_wp) .or. &
         .not. is_close(grad_of_G_subtracted%re, -0.05811018921675381_wp) .or. &
         .not. is_close(grad_of_G_subtracted%im, -0.01780306373513057_wp) ) then
       msg = 'Test 2 failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if

    ! Test 3
    R = 0._wp
    kR = wavenumber*R
    call green_func_subtracted(&
         G_subtracted, &
         grad_of_G_subtracted, &
         wavenumber          , &
         kR                  , &
         R                   , &
         faces_are_on_same_plane)
    if ( .not. is_close(G_subtracted%re, 0._wp) .or. &
         .not. is_close(G_subtracted%im, 0._wp) .or. &
         .not. is_close(grad_of_G_subtracted%re, 0._wp) .or. &
         .not. is_close(grad_of_G_subtracted%im, 0._wp) ) then
       msg = 'Test 3 failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if

    ! Test 4
    wavenumber = cmplx(0._wp, 0._wp)
    R = 3._wp
    kR = wavenumber*R
    call green_func_subtracted(&
         G_subtracted, &
         grad_of_G_subtracted, &
         wavenumber          , &
         kR                  , &
         R                   , &
         faces_are_on_same_plane)
    if ( .not. is_close(G_subtracted%re, 0._wp) .or. &
         .not. is_close(G_subtracted%im, 0._wp) .or. &
         .not. is_close(grad_of_G_subtracted%re, 0._wp) .or. &
         .not. is_close(grad_of_G_subtracted%im, 0._wp) ) then
       msg = 'Test 4 failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    
    call print_status_done(test_nr)
  end subroutine test_green_func_subtracted
  
  !!----------------------------------------------------------------------------
  
  subroutine test_eval_subtracted_terms(num_tests, test_nr, PMCHW)
    integer     , intent(in)          :: num_tests
    integer     , intent(inout)       :: test_nr
    type (PMCHW_RWG_type), intent(in) :: PMCHW
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    
    test_name = 'eval_subtracted_terms'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    ! Testing this function is difficult without having a code to reference the
    ! analytical calculations of the integrals over certain triangles.
    ! References usually plot error relative to the same code's results using
    ! a "sufficient" amount of quadrature points.
    !
    ! One solution may be to do a convergence test, as the results should
    ! converge with increasing amount of quadrature points.
    ! Another is to make sure the code looks correct, and save the results,
    ! so that at least an change to the function would be noticable.

    call print_status_done(test_nr)
    
  end subroutine test_eval_subtracted_terms
 
  !!----------------------------------------------------------------------------
  
  subroutine test_dbl_singularity_intgr(num_tests, test_nr, PMCHW)
    integer     , intent(in)          :: num_tests
    integer     , intent(inout)       :: test_nr
    type (PMCHW_RWG_type), intent(in) :: PMCHW
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    
    test_name = 'dbl_singularity_intgr'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    ! Testing this function is difficult without having a code to reference the
    ! analytical calculations of the integrals over certain triangles.
    ! References usually plot error relative to the same code's results using
    ! a "sufficient" amount of quadrature points.
    !
    ! One solution may be to do a convergence test, as the results should
    ! converge with increasing amount of quadrature points.
    ! Another is to make sure the code looks correct, and save the results,
    ! so that at least an change to the function would be noticable.

    call print_status_done(test_nr)
    
  end subroutine test_dbl_singularity_intgr
  
  !!----------------------------------------------------------------------------
  
  subroutine test_get_edge_lengths(num_tests, test_nr, PMCHW)
    integer              , intent(in)    :: num_tests
    integer              , intent(inout) :: test_nr
    type (PMCHW_RWG_type), intent(in)    :: PMCHW
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable                           :: test_name 
    character(len=255)                                  :: msg
    integer   , parameter                               :: SPATIAL_DIM = 3
    integer   , parameter                               :: NUM_FACE_VERTICES = 3
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: face_coords
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: check1
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: check2
    real(wp), dimension(2, SPATIAL_DIM)                 :: edge_coords
    real(wp), dimension(SPATIAL_DIM)                    :: res1
    real(wp), dimension(SPATIAL_DIM)                    :: res2
    integer , dimension(NUM_FACE_VERTICES)              :: face_edges
    integer                                             :: p
    integer                                             :: l
    
    test_name = 'get_edge_lengths'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

!!$    print *, ''
!!$    do p = 1, PMCHW%RWG_basis%mesh%num_faces
!!$       print *, p
!!$       face_edges = PMCHW%RWG_basis%mesh%get_edges_on_face(p)
!!$       face_coords = PMCHW%RWG_basis%mesh%get_face_coords(p)
!!$       do l = 1, NUM_FACE_VERTICES
!!$          edge_coords = PMCHW%RWG_basis%mesh%get_edge_coords(face_edges(l))
!!$          print *, '    ', face_coords(l, :), edge_coords(1, :), edge_coords(2, :)
!!$       end do
!!$    end do
          
    res1 = PMCHW%get_edge_lengths(1)
    res2 = PMCHW%get_edge_lengths(8)
    check1 = PMCHW%RWG_basis%mesh%get_face_coords(1)
    check2 = PMCHW%RWG_basis%mesh%get_face_coords(8)
    if ( .not. all(is_close(res1, [ 1._wp, sqrt(2._wp), 1._wp ])) .or. &
         .not. all(is_close(res2, [ 1._wp, 1._wp, sqrt(2._wp) ])) .or. &
         .not. all(is_close(check1(1, :), [ 0._wp, 0._wp, 0._wp ])) .or. & 
         .not. all(is_close(check1(2, :), [ 1._wp, 0._wp, 0._wp ])) .or. & 
         .not. all(is_close(check1(3, :), [ 1._wp, 0._wp, 1._wp ])) .or. & 
         .not. all(is_close(check2(1, :), [ 0._wp, 1._wp, 1._wp ])) .or. & 
         .not. all(is_close(check2(2, :), [ 1._wp, 0._wp, 1._wp ])) .or. & 
         .not. all(is_close(check2(3, :), [ 1._wp, 1._wp, 1._wp ])) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
          
    call print_status_done(test_nr)
  end subroutine test_get_edge_lengths
  
  !!----------------------------------------------------------------------------
  
  subroutine test_face_centroid(num_tests, test_nr, PMCHW)
    integer              , intent(in)    :: num_tests
    integer              , intent(inout) :: test_nr
    type (PMCHW_RWG_type), intent(in)    :: PMCHW
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable        :: test_name 
    character(len=255)               :: msg
    integer   , parameter            :: SPATIAL_DIM = 3
    integer   , parameter            :: NUM_FACE_VERTICES = 3
    real(wp), dimension(SPATIAL_DIM) :: res1
    real(wp), dimension(SPATIAL_DIM) :: res2
    real(wp), dimension(SPATIAL_DIM) :: res3
    real(wp), dimension(SPATIAL_DIM) :: res4
    real(wp), dimension(SPATIAL_DIM) :: correct1
    real(wp), dimension(SPATIAL_DIM) :: correct2
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: intgr_pnts1
    real(wp), dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: intgr_pnts2
    
    test_name = 'face_centroid'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    intgr_pnts1(1, :) =  &
         [ 0._wp, 0._wp, 0._wp ]*GQF_triangle_4pnt(1, 2) &
         + [ 1._wp, 0._wp, 0._wp ]*GQF_triangle_4pnt(1, 3) &
         + [ 1._wp, 0._wp, 1._wp ]*GQF_triangle_4pnt(1, 4)
    intgr_pnts2(1, :) =  &
         [ 1._wp, 1._wp, 0._wp ]*GQF_triangle_4pnt(1, 2) &
         + [ 1._wp, 0._wp, 1._wp ]*GQF_triangle_4pnt(1, 3) &
         + [ 1._wp, 0._wp, 0._wp ]*GQF_triangle_4pnt(1, 4)
    res1 = PMCHW%face_centroid(1, 1, intgr_pnts1)
    res2 = PMCHW%face_centroid(1, 4, intgr_pnts1)
    res3 = PMCHW%face_centroid(12, 1, intgr_pnts2)
    res4 = PMCHW%face_centroid(12, 4, intgr_pnts2)
    correct1 = [ 2._wp/3._wp, 0._wp, 1._wp/3._wp ]
    correct2 = [ 1._wp, 1._wp/3._wp, 1._wp/3._wp ]
    if ( .not. all(is_close(res1, correct1)) .or. &
         .not. all(is_close(res2, correct1)) .or. &
         .not. all(is_close(res3, correct2)) .or. &
         .not. all(is_close(res4, correct2)) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    call print_status_done(test_nr)
  end subroutine test_face_centroid
  
  !!----------------------------------------------------------------------------
    
  subroutine test_are_obs_pnt_and_src_close(num_tests, test_nr, PMCHW)
    integer              , intent(in)    :: num_tests
    integer              , intent(inout) :: test_nr
    type (PMCHW_RWG_type), intent(in)    :: PMCHW
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable        :: test_name 
    character(len=255)               :: msg
    complex(wp)                      :: wavenumber
    integer                          :: p
    integer                          :: q
    logical                          :: res1
    logical                          :: res2
    logical                          :: res3
    
    test_name = 'are_obs_pnt_and_src_close'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)
    ! Assuming PROP_CONST_OBS_PNT_SRC_CLOSE <= 0
    wavenumber = 0._wp
    p = 1
    q = 1
    res1 = PMCHW%are_obs_pnt_and_src_close(&
         PMCHW%RWG_basis%mesh%face_centroid(p), &
         PMCHW%RWG_basis%mesh%face_centroid(q), &
         wavenumber                           , &
         p=p                                  , &
         q=q)
    q = 2
    res2 = PMCHW%are_obs_pnt_and_src_close(&
         PMCHW%RWG_basis%mesh%face_centroid(p), &
         PMCHW%RWG_basis%mesh%face_centroid(q), &
         wavenumber                           , &
         p=p                                  , &
         q=q)
    q = 9
    res3 = PMCHW%are_obs_pnt_and_src_close(&
         PMCHW%RWG_basis%mesh%face_centroid(p), &
         PMCHW%RWG_basis%mesh%face_centroid(q), &
         wavenumber                           , &
         p=p                                  , &
         q=q)
    if ( .not. (res1 .eqv. .true.) .or. &
         .not. (res2 .eqv. .true.) .or. &
         .not. (res3 .eqv. .false.) ) then
       msg = 'Tests failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    call print_status_done(test_nr)
  end subroutine test_are_obs_pnt_and_src_close
  
  !!----------------------------------------------------------------------------
  
end program main
