!!==============================================================================
! This program runs tests for the module 'RWG_basis'.
!
!   First a series of subroutines are defines for testing specific parts of 
!   the module. Then a sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: October 28th 2020.
!!==============================================================================


!!===========!!
! Subroutines !
!!============!=================================================================
subroutine test_initialise(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: check_RWG_basis_initialisation
  use mesh_mod         , only: mesh_type
  use RWG_basis_mod    , only: RWG_basis_type
  implicit none
  integer, intent(in)                    :: num_tests
  integer, intent(inout)                 :: test_nr
  ! Variables for internal use -------------------------------------------------
  type (mesh_type)                       :: mesh
  type (RWG_basis_type)                  :: RWG_basis
  character(len=:), allocatable          :: filename
  character(len=:), allocatable          :: file_format
  logical                                :: closed_surface
  integer                                :: edge_order
  integer                                :: spatial_dim
  integer, parameter                     :: num_face_vertices = 3
  integer                                :: num_handles
  integer                                :: num_apertures
  integer                                :: num_boundary_edges
  integer                                :: correct_num_bases
  integer , dimension(:)   , allocatable :: correct_basis_edges
  integer , dimension(:, :), allocatable :: correct_adjacencies
  real(wp), dimension(:)   , allocatable :: correct_basis_edge_lengths
  integer                                :: i
  integer                                :: j
  
  call print_status_start(num_tests, test_nr, 'RWG_basis_mod', &
       'initialise')
  print *, ''
  !!----------------!!
  ! Test tetrahedron !
  !------------------!----------------------------------------------------------
  print *, 'Tetrahedron ...'
  filename = 'meshes/tetrahedron.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0
  correct_num_bases = 6
  allocate(correct_basis_edges(correct_num_bases))
  allocate(correct_adjacencies(correct_num_bases, 2))
  allocate(correct_basis_edge_lengths(correct_num_bases))
  correct_basis_edges = [ 1, 2, 3, 4, 5, 6 ]
  correct_adjacencies(:, 1) = [ 1, 1, 1, 2, 2, 3 ]
  correct_adjacencies(:, 2) = [ 4, 3, 2, 3, 4, 4 ]

  do i = 1, correct_num_bases
     correct_basis_edge_lengths(i) = 2._wp*sqrt(2._wp)
  end do
       
  call mesh%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call RWG_basis%initialise(mesh)
  call check_RWG_basis_initialisation(&
       RWG_basis          , &
       correct_basis_edges, &
       correct_adjacencies, &
       correct_basis_edge_lengths)
  
  !!----------------!!
  ! Test closed cube !
  !------------------!----------------------------------------------------------
  print *, 'Closed cube ...'
  filename = 'meshes/cube_closed.msh'
  file_format = 'gmsh2'
  correct_num_bases = 18
  deallocate(correct_basis_edges)
  deallocate(correct_adjacencies)
  deallocate(correct_basis_edge_lengths)
  allocate(correct_basis_edges(correct_num_bases))
  allocate(correct_adjacencies(correct_num_bases, 2))
  allocate(correct_basis_edge_lengths(correct_num_bases))
  correct_basis_edges = [ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, &
       16, 17, 18 ]
  correct_adjacencies(:, 1) = [ 1,  1, 1, 2, 2, 3, 3, 4,  4, 5,  5,  6, 7, &
        8, 8,  9,  9, 11]
  correct_adjacencies(:, 2) = [ 6, 12, 2, 7, 3, 7, 4, 10, 5, 10, 6, 12, 8, &
       11, 9, 11, 10, 12]

  do i = 1, correct_num_bases
     if (any([3, 7, 11, 13, 17, 18 ] == i)) then
        correct_basis_edge_lengths(i) = sqrt(2._wp)
     else
        correct_basis_edge_lengths(i) = 1._wp
     end if
  end do
       
  call mesh%deallocate_attributes()
  call mesh%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call RWG_basis%deallocate_attributes()
  call RWG_basis%initialise(mesh)
  call check_RWG_basis_initialisation(&
       RWG_basis          , &
       correct_basis_edges, &
       correct_adjacencies, &
       correct_basis_edge_lengths)
   
  !!----------------!!
  ! Test open cube   !
  !------------------!----------------------------------------------------------
!!$  print *, 'Open cube ...'
!!$  filename = 'meshes/cube_open.msh'
!!$  file_format = 'gmsh2'
!!$  closed_surface = .false.
!!$  num_apertures = 1
!!$  num_boundary_edges = 3
!!$  correct_num_bases = 15
!!$  deallocate(correct_basis_edges)
!!$  deallocate(correct_adjacencies)
!!$  deallocate(correct_basis_edge_lengths)
!!$  allocate(correct_basis_edges(correct_num_bases))
!!$  allocate(correct_adjacencies(correct_num_bases, 2))
!!$  allocate(correct_basis_edge_lengths(correct_num_bases))
!!$  correct_basis_edges = [ 1, 3, 4, 5, 6, 7, 8, 9, 10, 11, 13, 14, 15, 16, 17]
!!$  correct_adjacencies(:, 1) = [ 1, 1, 2, 2, 3, 3, 4,  4, 5,  5, 7, 8, 8,  9,  9]
!!$  correct_adjacencies(:, 2) = [ 6, 2, 7, 3, 7, 4, 10, 5, 10, 6, 8,11, 9, 11, 10]
!!$
!!$  do i = 1, correct_num_bases
!!$     if (any([2, 6, 10, 11, 15] == i)) then
!!$        correct_basis_edge_lengths(i) = sqrt(2._wp)
!!$     else
!!$        correct_basis_edge_lengths(i) = 1._wp
!!$     end if
!!$  end do
!!$       
!!$  call mesh%deallocate_attributes()
!!$  call mesh%initialise(&
!!$       filename                      , &
!!$       file_format                   , &
!!$       closed_surface                , &
!!$       edge_order                    , &
!!$       spatial_dim                   , &
!!$       num_face_vertices             , &
!!$       num_handles_in=num_handles    , &
!!$       num_apertures_in=num_apertures, &
!!$       num_boundary_edges_in=num_boundary_edges)
!!$  call RWG_basis%deallocate_attributes()
!!$  call RWG_basis%initialise(mesh)
!!$  call check_RWG_basis_initialisation(&
!!$       RWG_basis          , &
!!$       correct_basis_edges, &
!!$       correct_adjacencies, &
!!$       correct_basis_edge_lengths)
!!$
!!$  !!----------------!!
!!$  ! Test 3D triangle !
!!$  !------------------!----------------------------------------------------------
!!$  print *, '3D triangle ...'
!!$  filename = 'meshes/triangle_3D.msh'
!!$  file_format = 'gmsh2'
!!$  closed_surface = .false.
!!$  num_apertures = 1
!!$  num_boundary_edges = 3
!!$  correct_num_bases = 0
!!$
!!$  call mesh%deallocate_attributes()
!!$  call mesh%initialise(&
!!$       filename                      , &
!!$       file_format                   , &
!!$       closed_surface                , &
!!$       edge_order                    , &
!!$       spatial_dim                   , &
!!$       num_face_vertices             , &
!!$       num_handles_in=num_handles    , &
!!$       num_apertures_in=num_apertures, &
!!$       num_boundary_edges_in=num_boundary_edges)
!!$  call RWG_basis%deallocate_attributes()
!!$  call RWG_basis%initialise(mesh)
!!$  if (RWG_basis%num_bases /= correct_num_bases) then
!!$     print *, 'Error: test_initialise in RWG_basis_test.f90 failed..'
!!$     print *, '       num_bases:', RWG_basis%num_bases, &
!!$          ' Correct:', correct_num_bases
!!$     stop 2
!!$  end if
!!$
!!$  !!----------------!!
!!$  ! Test 3D square   !
!!$  !------------------!----------------------------------------------------------
!!$  print *, '3D square ...'
!!$  filename = 'meshes/square_3D.msh'
!!$  file_format = 'gmsh2'
!!$  closed_surface = .false.
!!$  num_apertures = 1
!!$  num_boundary_edges = 4
!!$  correct_num_bases = 1
!!$  deallocate(correct_basis_edges)
!!$  deallocate(correct_adjacencies)
!!$  deallocate(correct_basis_edge_lengths)
!!$  allocate(correct_basis_edges(correct_num_bases))
!!$  allocate(correct_adjacencies(correct_num_bases, 2))
!!$  allocate(correct_basis_edge_lengths(correct_num_bases))
!!$  correct_basis_edges = [ 3 ]
!!$  correct_adjacencies(:, 1) = [ 1 ]
!!$  correct_adjacencies(:, 2) = [ 2 ]
!!$  correct_basis_edge_lengths(1) = sqrt(2._wp)
!!$
!!$  call mesh%deallocate_attributes()
!!$  call mesh%initialise(&
!!$       filename                      , &
!!$       file_format                   , &
!!$       closed_surface                , &
!!$       edge_order                    , &
!!$       spatial_dim                   , &
!!$       num_face_vertices             , &
!!$       num_handles_in=num_handles    , &
!!$       num_apertures_in=num_apertures, &
!!$       num_boundary_edges_in=num_boundary_edges)
!!$  call RWG_basis%deallocate_attributes()
!!$  call RWG_basis%initialise(mesh)
!!$  call check_RWG_basis_initialisation(&
!!$       RWG_basis          , &
!!$       correct_basis_edges, &
!!$       correct_adjacencies, &
!!$       correct_basis_edge_lengths)
!!$  
  !!----------------!!
  ! Test 2D triangle !
  !------------------!----------------------------------------------------------
!!$  print *, '2D triangle ...'
!!$  filename = 'meshes/triangle_2D.msh'
!!$  file_format = 'gmsh2'
!!$  spatial_dim = 2
!!$  closed_surface = .false.
!!$  num_apertures = 1
!!$  num_boundary_edges = 3
!!$  correct_num_bases = 0
!!$
!!$  call mesh%deallocate_attributes()
!!$  call mesh%initialise(&
!!$       filename                      , &
!!$       file_format                   , &
!!$       closed_surface                , &
!!$       edge_order                    , &
!!$       spatial_dim                   , &
!!$       num_face_vertices             , &
!!$       num_handles_in=num_handles    , &
!!$       num_apertures_in=num_apertures, &
!!$       num_boundary_edges_in=num_boundary_edges)
!!$  call RWG_basis%deallocate_attributes()
!!$  call RWG_basis%initialise(mesh)
!!$  if (RWG_basis%num_bases /= correct_num_bases) then
!!$     print *, 'Error: test_initialise in RWG_basis_test.f90 failed..'
!!$     print *, '       num_bases:', RWG_basis%num_bases, &
!!$          ' Correct:', correct_num_bases
!!$     stop 2
!!$  end if

  !!----------------!!
  ! Test 2D square   !
  !------------------!----------------------------------------------------------
!!$  print *, '2D square ...'
!!$  filename = 'meshes/square_2D.msh'
!!$  file_format = 'gmsh2'
!!$  spatial_dim = 2
!!$  closed_surface = .false.
!!$  num_apertures = 1
!!$  num_boundary_edges = 4
!!$  correct_num_bases = 1
!!$  deallocate(correct_basis_edges)
!!$  deallocate(correct_adjacencies)
!!$  deallocate(correct_basis_edge_lengths)
!!$  allocate(correct_basis_edges(correct_num_bases))
!!$  allocate(correct_adjacencies(correct_num_bases, 2))
!!$  allocate(correct_basis_edge_lengths(correct_num_bases, 2))
!!$  correct_basis_edges = [ 3 ]
!!$  correct_adjacencies(:, 1) = [ 1 ]
!!$  correct_adjacencies(:, 2) = [ 2 ]
!!$  correct_basis_edge_lengths(1, 1) = sqrt(2._wp)
!!$  correct_basis_edge_lengths(1, 2) = -sqrt(2._wp)
!!$
!!$  call mesh%deallocate_attributes()
!!$  call mesh%initialise(&
!!$       filename                      , &
!!$       file_format                   , &
!!$       closed_surface                , &
!!$       edge_order                    , &
!!$       spatial_dim                   , &
!!$       num_face_vertices             , &
!!$       num_handles_in=num_handles    , &
!!$       num_apertures_in=num_apertures, &
!!$       num_boundary_edges_in=num_boundary_edges)
!!$  print *, 'deallo RWG..'
!!$  call RWG_basis%deallocate_attributes()
!!$  print *, 'init RWG..'
!!$  call RWG_basis%initialise(mesh)
!!$  print *, 'Checking RWG..'
!!$  call check_RWG_basis_initialisation(&
!!$       RWG_basis          , &
!!$       correct_basis_edges, &
!!$       correct_adjacencies, &
!!$       correct_basis_edge_lengths)
!!$
  call print_status_done(test_nr)
end subroutine test_initialise

!!------------------------------------------------------------------------------

subroutine test_integrate_tested_func(num_tests, test_nr)
  use working_precision  , only: wp
  use test_utilities     , only: print_status_start
  use test_utilities     , only: print_status_done
  use is_close_mod       , only: is_close
  use mesh_mod           , only: mesh_type
  use RWG_basis_mod      , only: RWG_basis_type
  use gauss_quad_formulas_mod, only: GQF_triangle_1pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use math_funcs_mod         , only: standard_basis
  use math_funcs_mod         , only: scale_with_1
  use math_funcs_mod         , only: first_component

  implicit none
  integer, intent(in)                    :: num_tests
  integer, intent(inout)                 :: test_nr
  ! Variables for internal use -------------------------------------------------
  type (mesh_type)                       :: mesh
  type (RWG_basis_type)                  :: RWG_basis
  character(len=:), allocatable          :: filename
  character(len=:), allocatable          :: file_format
  logical                                :: closed_surface
  integer                                :: edge_order
  integer                                :: spatial_dim
  integer, parameter                     :: num_face_vertices = 3
  integer                                :: num_handles
  integer                                :: num_apertures
  integer                                :: num_boundary_edges
  real(wp), dimension(:), allocatable :: integration_values
  real(wp), dimension(:), allocatable :: correct_intgr_values
  real(wp), parameter                 :: relative_tol = 1e-7_wp
  integer                             :: i
  
  call print_status_start(num_tests, test_nr, 'RWG_basis_mod', &
       'integrate_tested_func')

  !!----------------!!
  ! Test tetrahedron !
  !------------------!----------------------------------------------------------
  filename = 'meshes/tetrahedron.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0
  allocate(correct_intgr_values(6))

  call mesh%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call RWG_basis%initialise(mesh)

  allocate(integration_values(RWG_basis%num_bases))

  ! func: standard basis f(x) = 1
  correct_intgr_values = [ -8._wp/3._wp*sqrt(2._wp), -8._wp/3._wp*sqrt(2._wp), &
       -8._wp/3._wp*sqrt(2._wp), 0._wp, 0._wp, 0._wp ] 
  integration_values = RWG_basis%integrate_tested_func(&
       GQF_triangle_3pnt, &
       standard_basis)
 
  do i = 1, size(correct_intgr_values)
     if (.not. is_close(correct_intgr_values(i), integration_values(i), &
          rtol=relative_tol)) then
        print *, 'Error:'
        print *, '  Integration values differ: i =', i
        print *, '    Estimated:', integration_values(i)
        print *, '    Correct  :', correct_intgr_values(i)
     end if
  end do

  ! func: f(x) = x
  correct_intgr_values = [ 0._wp, 0._wp, 0._wp, 0._wp, 0._wp, 0._wp ]
  integration_values = RWG_basis%integrate_tested_func(&
       GQF_triangle_3pnt, &
       scale_with_1)
  do i = 1, size(correct_intgr_values)
     if (.not. is_close(correct_intgr_values(i), integration_values(i), &
          rtol=relative_tol)) then
        print *, 'Error:'
        print *, '  Integration values differ: i =', i
        print *, '    Estimated:', integration_values(i)
        print *, '    Correct  :', correct_intgr_values(i)
     end if
  end do


  ! func: f(x) = [x_1, 0, 0]
  integration_values = RWG_basis%integrate_tested_func(&
       GQF_triangle_3pnt, &
       first_component)
 
  call print_status_done(test_nr)
end subroutine test_integrate_tested_func

!!============!!
! Main program !
!!=============!================================================================
program main
  use test_utilities, only: print_status_start
  use test_utilities, only: print_status_done
  use test_utilities, only: print_error_message
  use RWG_basis_mod , only: RWG_basis_type
  use mesh_mod      , only: mesh_type

  implicit none
  character(:), allocatable :: module_name 
  integer     , parameter   :: num_tests = 3
  integer                   :: test_nr = 1
  ! Variables to create RWG_basis_type ojects ----------------------------------
  type (mesh_type)          :: mesh2
  type (mesh_type)          :: mesh3
  type (RWG_basis_type)     :: closed_cube
  type (RWG_basis_type)     :: sphere
  character(:), allocatable :: filename
  character(:), allocatable :: file_format
  integer                   :: edge_order
  integer                   :: spatial_dim
  integer                   :: face_order
  integer                   :: num_handles
  integer                   :: num_apertures
  integer                   :: num_boundary_edges
  logical                   :: closed_surface

  module_name = 'RWG_basis_mod'

  filename = 'meshes/cube_closed.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  face_order = 3
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0
  call mesh2%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       face_order                    , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call closed_cube%initialise(mesh2)
  filename = 'meshes/gmsh_sphere_1ord.msh'
  call mesh3%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       face_order                    , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call sphere%initialise(mesh3)
 
  
  call test_initialise(num_tests, test_nr)
  call test_integrate_tested_func(num_tests, test_nr)
  call test_write_RWG_basis(num_tests, test_nr, sphere)

  !=======!=========================!==========================================!
contains  ! /\/\/\/\/\/\/\/\/\/\/\/\!/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\!
  !=======!=========================!==========================================!
  
!!===========!!
! Subroutines !
!!============!=================================================================
  subroutine test_write_RWG_basis(num_tests, test_nr, RWG_basis)
    integer     , intent(in)          :: num_tests
    integer     , intent(inout)       :: test_nr
    type (RWG_basis_type), intent(in) :: RWG_basis
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable                   :: test_name 
    character(len=255)                          :: msg
    character(len=255)                          :: filename
 
    test_name = 'write_RWG_basis'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    filename = 'RWG_basis.txt'
    call RWG_basis%write_RWG_basis(filename)


!!$    if (error_nr /= 0) then
!!$       write (msg, "(A5, I2, A8)") 'Test ', error_nr, ' failed.'
!!$       call print_error_message(module_name, test_name, msg, degree=2)
!!$    end if
       
    call print_status_done(test_nr)
  end subroutine test_write_RWG_basis
  
  !!----------------------------------------------------------------------------
  
end program main
