!!==============================================================================
! This program runs tests for the module 'io_mod'.
!
!   First a series of subroutines are defines for testing specific parts of 
!   the module. Then a sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: October 21th 2020.
!!==============================================================================


!!===========!!
! Subroutines !
!!============!=================================================================
subroutine test_capitalise_char(num_tests, test_nr)
  use test_utilities, only: print_status_start, print_status_done
  use io_mod        , only: capitalise_char
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  character(len=255)     :: text
  character(len=255)     :: correct_result
  integer                :: i

  call print_status_start(num_tests, test_nr, 'io_mod', 'capitalise_char')

  text = 'abczABCZ 093-.,eEdD'
  correct_result = 'ABCZABCZ 093-.,EEDD'
  do i = 1, len_trim(text)
     text(i:i) = capitalise_char(text(i:i))
  end do
  if (text /= correct_result) then
     print *, 'Error: test_capitalise_char() in io_test.f90 failed..'
     stop 1
  end if

  call print_status_done(test_nr)
end subroutine test_capitalise_char

!!------------------------------------------------------------------------------

subroutine test_string_to_int4(num_tests, test_nr)
  use test_utilities, only: print_status_start, print_status_done
  use io_mod        , only: string_to_int4
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  character(len=255)     :: text
  integer, dimension(8)  :: correct_result
  integer                :: int4_value
  integer                :: length
  integer                :: error_state
  integer                :: i

  call print_status_start(num_tests, test_nr, 'io_mod', 'test_string_to_int4')

  text = '66 0 1 9 2340 -2 -3333 +99 .'
  correct_result = [ 66, 0, 1, 9, 2340, -2, -3333, 99 ]
  length = 1
  do i = 1, size(correct_result)
     call string_to_int4(text, length, int4_value, error_state)
     text = text(length + 1:)
     if (int4_value /= correct_result(i)) then
        print *, 'Error: test_string_to_real_wp() in io_test.f90 failed..'
        print *, 'i: ', i, '    Value: ', int4_value
        stop 1
     end if
  end do

  call print_status_done(test_nr)
end subroutine test_string_to_int4

!!------------------------------------------------------------------------------

subroutine test_string_to_real_wp(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use io_mod           , only: string_to_real_wp
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  character(len=255)     :: text 
  real(wp), dimension(4) :: correct_result
  real(wp)               :: value_of_real
  integer                :: length
  integer                :: error_state
  integer                :: i
  real(wp), parameter    :: relative_tol = 1e-7_wp

  call print_status_start(num_tests, test_nr, 'io_mod', 'string_to_real_wp')

  text = '0 12.3456 3.9e-3 -3.0'
  correct_result = [ real(0, wp), real(12.3456, wp), &
       real(0.0039, wp), real(-3.0, wp) ] 
  length = 1
  do i = 1, size(correct_result)
     call string_to_real_wp(text, length, value_of_real, error_state)
     text = text(length + 1:)
     if (.not. is_close(value_of_real, correct_result(i), &
          rtol=relative_tol)) then
        print *, ''
        print *, 'Error: test_string_to_real_wp() in io_test.f90 failed..'
        print *, 'i: ', i, '    Value: ', value_of_real, 'Desired: ', &
             correct_result(i)
        stop 1
     end if
  end do

  call print_status_done(test_nr)
end subroutine test_string_to_real_wp

!-------------------------------------------------------------------------------

subroutine test_count_int4_on_string(num_tests, test_nr)
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use io_mod           , only: count_int4_on_string
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  character(len=255)     :: text 
  integer                :: correct_result
  integer                :: num_int4
  
  call print_status_start(num_tests, test_nr, 'io_mod', 'count_int4_on_string')

  text = '0 -1 223 -32 00'
  correct_result = 5
  num_int4 = count_int4_on_string(text)
  if (num_int4 /= correct_result) then
     print *, ''
     print *, 'Error: test_count_int4_on_string in io_test.f90 failed..'
     print *, '       Value: ', num_int4, 'Desired: ', correct_result
     stop 1
  end if

  call print_status_done(test_nr)
end subroutine test_count_int4_on_string

!!------------------------------------------------------------------------------

subroutine test_count_real_wp_on_string(num_tests, test_nr)
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use io_mod           , only: count_real_wp_on_string
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  character(len=255)     :: text 
  integer                :: correct_result
  integer                :: num_real_wp
  
  call print_status_start(num_tests, test_nr, 'io_mod', &
       'count_real_wp_on_string')

  text = '0 -1.0 3e3 -2.2e-2.3 -0.0'
  correct_result = 5
  num_real_wp = count_real_wp_on_string(text)
  if (num_real_wp /= correct_result) then
     print *, ''
     print *, 'Error: test_count_real_wp_on_string in io_test.f90 failed..'
     print *, '       Value: ', num_real_wp, 'Desired: ', correct_result
     stop 1
  end if

  call print_status_done(test_nr)
end subroutine test_count_real_wp_on_string

!!------------------------------------------------------------------------------

subroutine test_read_nth_int4(num_tests, test_nr)
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use io_mod           , only: read_nth_int4
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  character(len=255)     :: text 
  integer                :: int4_value
  integer, dimension(3)  :: correct_result
  integer, dimension(3)  :: nth
  integer                :: i
  
  call print_status_start(num_tests, test_nr, 'io_mod', 'read_nth_int4')

  text = '0 -1 223 -32 00 23 112000'
  nth = [ 1, 7, 3 ]
  correct_result = [ 0, 112000, 223 ]
  do i = 1, size(nth)
     int4_value = read_nth_int4(text, nth(i))
     if (int4_value /= correct_result(i)) then
        print *, ''
        print *, 'Error: test_read_nth_int4 in io_test.f90 failed..'
        print *, '       Value: ',  int4_value, 'Desired: ', correct_result(i)
        stop 1
     end if
  end do
  
  call print_status_done(test_nr)
end subroutine test_read_nth_int4

!!------------------------------------------------------------------------------

subroutine test_read_n_last_int4(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start, print_status_done
  use io_mod           , only: read_n_last_int4
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  character(len=255)     :: text
  integer, parameter     :: n = 2
  integer, dimension(n)  :: n_last_int4
  integer, dimension(n)  :: correct_result
  integer                :: i

  call print_status_start(num_tests, test_nr, 'io_mod', 'read_n_last_int4')

  text = '-1 93 00 -44'
  correct_result = [ 0, -44 ]
  call read_n_last_int4(text, n, n_last_int4)
  do i = 1, n
     if (n_last_int4(i) /= correct_result(i)) then
        print *, 'Error: test_read_n_last_int4() in io_test.f90 failed..'
        print *, 'i: ', i, '    Value: ', n_last_int4(i)
        stop 1
     end if
  end do
  
  call print_status_done(test_nr)
end subroutine test_read_n_last_int4

!!------------------------------------------------------------------------------

subroutine test_read_n_last_real_wp(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use is_close_mod     , only: is_close
  use io_mod           , only: read_n_last_real_wp

  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  ! Variables for internal use -------------------------------------------------
  character(len=255)     :: text
  integer , parameter    :: n = 2
  real(wp), dimension(n) :: n_last_real_wp
  real(wp), dimension(n) :: correct_result
  integer                :: i
  real(wp), parameter    :: relative_tol = 1e-7_wp
  
  call print_status_start(num_tests, test_nr, 'io_mod', 'read_n_last_wp')

  text = '-1. 93.213 00 -4.3e-2'
  correct_result = [ real(0, wp), real(-0.043, wp) ]
  call read_n_last_real_wp(text, n, n_last_real_wp)
  do i = 1, n
     if (.not. is_close(n_last_real_wp(i), correct_result(i), &
          rtol=relative_tol)) then
        print *, ''
        print *, 'Error: test_read_n_last_real_wp() in io_test.f90 failed..'
        print *, 'i: ', i, '    Value: ', n_last_real_wp(i), 'Desired: ', &
             correct_result(i)
        stop 1
     end if
  end do 

  call print_status_done(test_nr)
end subroutine test_read_n_last_real_wp

!!------------------------------------------------------------------------------

subroutine test_open_read_gmsh2(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start, print_status_done
  use io_mod           , only: open_read_gmsh2
  implicit none
  integer, intent(in)                               :: num_tests
  integer, intent(inout)                            :: test_nr
  character(len=255)                                :: MSH_FILE
  character, dimension(:, :), allocatable           :: msh_data
  integer  , parameter                              :: spatial_dim = 3
  integer  , parameter                              :: element_order = 3
  integer  , parameter                              :: num_nodes = 4
  integer  , parameter                              :: num_elements = 4
  real(wp) , dimension(:, :), allocatable           :: nodes
  integer  , dimension(:, :), allocatable           :: elements
  real(wp) , dimension(num_nodes, spatial_dim)      :: correct_nodes
  integer  , dimension(num_elements, element_order) :: correct_elements
  integer                                           :: i
  integer                                           :: j

  call print_status_start(num_tests, test_nr, 'io_mod', 'open_read_gmsh2')
  
  MSH_FILE = 'meshes/tetrahedron.msh'
  correct_nodes(1, :) = [ 1._wp, -1._wp, 1._wp ]
  correct_nodes(2, :) = [ -1._wp, 1._wp, 1._wp ]
  correct_nodes(3, :) = [ 1._wp, 1._wp, -1._wp ]
  correct_nodes(4, :) = [ -1._wp, -1._wp, -1._wp ]
  correct_elements(1, :) = [ 1, 3, 2 ]
  correct_elements(2, :) = [ 1, 2, 4 ]
  correct_elements(3, :) = [ 2, 3, 4 ]
  correct_elements(4, :) = [ 1, 4, 3 ]

  call open_read_gmsh2(MSH_FILE, spatial_dim, element_order, nodes, elements)

  do i = 1, num_nodes
     do j = 1, spatial_dim
        if (nodes(i, j) /= correct_nodes(i, j)) then
           print *, 'Error: test_open_read_gmsh2() in io_test.f90 failed..'
           print *, 'i: ', i, 'j: ', j, ' Node value: ', nodes(i, j)
           stop 1
        end if
     end do
  end do
  
  do i = 1, num_elements
     do j = 1, element_order
        if (elements(i, j) /= correct_elements(i, j)) then
           print *, 'Error: test_open_read_gmsh2() in io_test.f90 failed..'
           print *, 'i: ', i, 'j: ', j, ' Element value: ', elements(i, j)
           stop 1
        end if
     end do
  end do
    
  call print_status_done(test_nr)
end subroutine test_open_read_gmsh2


!!============!!
! Main program !
!!=============!================================================================
program main

  implicit none
  integer, parameter :: num_tests = 9
  integer            :: test_nr = 1
  call test_capitalise_char(num_tests, test_nr)
  call test_string_to_int4(num_tests, test_nr)
  call test_string_to_real_wp(num_tests, test_nr)
  call test_count_int4_on_string(num_tests, test_nr)
  call test_count_real_wp_on_string(num_tests, test_nr)
  call test_read_nth_int4(num_tests, test_nr)
  call test_read_n_last_int4(num_tests, test_nr)
  call test_read_n_last_real_wp(num_tests, test_nr)
  call test_open_read_gmsh2(num_tests, test_nr)
  ! Test r8mat_write
  ! Test get_unit
end program main

  
