!!============================================================================
! This program runs tests for the module 'is_close_mod'.
!
!   A series of subroutines are defines for testing specific parts of 
!   the module. A sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: January 17th 2020.
!!==============================================================================


!!=============!
! Main program !
!!=============!================================================================
program main
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: print_error_message
  use constants_mod    , only: I_IMAG
  use ieee_arithmetic  , only: ieee_support_inf
  use ieee_arithmetic  , only: ieee_value
  use ieee_arithmetic  , only: ieee_negative_inf
  use is_close_mod     , only: is_close

  implicit none
  character(:), allocatable :: module_name 
  integer     , parameter   :: num_tests = 1
  integer                   :: test_nr = 1
  module_name = 'is_close_mod'

  ! Call test routines
  call test_is_close(num_tests, test_nr)
  
  !=======!=========================!==========================================!
contains  ! /\/\/\/\/\/\/\/\/\/\/\/\!/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\!
  !=======!=========================!==========================================!

!!===========!!
! Subroutines !
!!============!=================================================================
  subroutine test_is_close(num_tests, test_nr)
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    ! Variables for internal use -----------------------------------------------
    complex(wp) :: exponential1
    complex(wp) :: exponential2
    real(wp)    :: r
    integer     :: error_id
    
    test_name = 'is_close'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    if (ieee_support_inf(r)) then
       r = ieee_value(r, ieee_negative_inf)
    end if
    
    exponential1 = exp(I_IMAG*0.5_wp)
    exponential2 = exp(I_IMAG*0.0_wp)

    ! Should be exactly equal
    if (.not. is_close(0._wp, 0._wp)) then
       error_id = 1
    ! Should be exactly equal
    else if (.not. is_close(1._wp, 1._wp)) then
       error_id = 2
    ! Should be approximately equal
    else if (.not. is_close(2._wp, 4._wp/2._wp)) then
       error_id = 3
    ! Should be approximately equal
    else if (.not. is_close(exponential1%re, cos(0.5_wp))) then
       error_id = 4
    ! Should terminate because of infinite value
    else if (is_close(0._wp, r)) then
       error_id = 5
    ! Should be approximately equal, but relative difference would be huge
    else if (.not. is_close(exponential2%im, sin(0._wp))) then
       error_id = 6
    ! Should be within default relative tolerance
    else if (.not. is_close(1._wp, (1._wp + 1.e-5))) then
       error_id = 7
    ! Should be outside default relative tolerance
    else if (is_close(1._wp, (1._wp + 1.e-4))) then
       error_id = 8
    ! Should be within default absolute tolerance
    else if (.not. is_close(0._wp, 99._wp*epsilon(0._wp))) then
       error_id = 9
    ! Should be outside default absolute tolerance
    else if (is_close(0._wp, 10001._wp*epsilon(0._wp))) then
       error_id = 10
    ! Testing an array
    else if (.not. all(is_close([0._wp, 1._wp, 2._wp], [0._wp, 1._wp, 2._wp])) &
         ) then
       error_id = 11
    else
       error_id = 0
    end if
    if (error_id /= 0) then
       write (msg, "(A5, I2, A8)") 'Test ', error_id, ' failed.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    
    call print_status_done(test_nr)
  end subroutine test_is_close

  !!----------------------------------------------------------------------------
    
end program main
