!!============================================================================
! This program runs tests for the module 'math_funcs_mod'.
!
!   A series of subroutines are defines for testing specific parts of 
!   the module. A sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: January 16th 2020.
!!==============================================================================


!!=============!
! Main program !
!!=============!================================================================
program main
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: print_error_message
  use is_close_mod     , only: is_close
  use math_funcs_mod   , only: cross_prod_3D
  use math_funcs_mod   , only: plane_wave

  implicit none
  character(:), allocatable :: module_name 
  integer     , parameter   :: num_tests = 2
  integer                   :: test_nr = 1
  module_name = 'math_funcs_mod'

  call test_cross_prod_3D(num_tests, test_nr)
  call test_plane_wave(num_tests, test_nr)
  
  !=======!=========================!==========================================!
contains  ! /\/\/\/\/\/\/\/\/\/\/\/\!/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\!
  !=======!=========================!==========================================!
  
!!===========!!
! Subroutines !
!!============!=================================================================
  subroutine test_cross_prod_3D(num_tests, test_nr)
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    character(:), allocatable            :: test_name 
    character(len=255)                   :: msg
    integer     , parameter              :: SPATIAL_DIM = 3
    real(wp)    , dimension(SPATIAL_DIM) :: vector_a
    real(wp)    , dimension(SPATIAL_DIM) :: vector_b
    real(wp)    , dimension(SPATIAL_DIM) :: vector_c
    real(wp)    , dimension(SPATIAL_DIM) :: vector_d
    real(wp)    , dimension(SPATIAL_DIM) :: vector_e
    real(wp)    , dimension(SPATIAL_DIM) :: vector_f
    
    test_name = 'cross_prod_3D'
    call print_status_start(num_tests, test_nr , module_name, &
         test_name)

    vector_a = [0._wp, 0._wp, 0._wp]
    vector_b = [1._wp, 0._wp, 0._wp]
    vector_c = [0._wp, 1._wp, 0._wp]
    vector_d = [0._wp, 0._wp, 1._wp]
    vector_e = [0.5_wp, 3._wp, 1.25_wp]
    vector_f = [1._wp, 2._wp, 3._wp]

    if (.not. all( &
         is_close(cross_prod_3D(vector_a, vector_a), [0._wp, 0._wp, 0._wp]))   &
         .or. .not. all( &
         is_close(cross_prod_3D(vector_b, vector_c), [0._wp, 0._wp, 1._wp]))   &
         .or. .not. all( &
         is_close(cross_prod_3D(vector_b, vector_d), [0._wp, -1._wp, 0._wp]))  &
         .or. .not. all( &
         is_close(cross_prod_3D(vector_c, vector_d), [1._wp, 0._wp, 0._wp]))   &
         .or. .not. all( &
         is_close(cross_prod_3D(vector_e, vector_f), [6.5_wp,-0.25_wp,-2._wp]))&
         .or. .not. all( &
         is_close(cross_prod_3D(vector_f, vector_e), [-6.5_wp,0.25_wp,2._wp])))&
         then
       msg = 'Incorrect test results.'
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    
    call print_status_done(test_nr)
  end subroutine test_cross_prod_3D

  !!----------------------------------------------------------------------------
    
  subroutine test_plane_wave(num_tests, test_nr)
    integer     , intent(in)    :: num_tests
    integer     , intent(inout) :: test_nr
    ! Variables for internal use -----------------------------------------------
    integer     , parameter     :: SPATIAL_DIM = 3
    character(:), allocatable   :: test_name 
    character(len=255)          :: msg
    complex(wp), dimension(SPATIAL_DIM) :: res1
    complex(wp), dimension(SPATIAL_DIM) :: res2
    complex(wp), dimension(SPATIAL_DIM) :: res3
    complex(wp), dimension(SPATIAL_DIM) :: amplitude
    complex(wp)                         :: wavenumber1
    complex(wp)                         :: wavenumber2
    real(wp)                            :: angular_frequency
    real(wp)                            :: time
    real(wp), dimension(SPATIAL_DIM)    :: r
    real(wp), dimension(SPATIAL_DIM)    :: wavevector

    test_name = 'plane_wave'
    call print_status_start(num_tests, test_nr, module_name, &
         test_name)

    wavenumber1 = cmplx(0.1_wp, 0._wp)
    wavenumber2 = cmplx(0.1_wp, 0.1_wp)
    angular_frequency = 0.1_wp
    time = 3._wp
    r = [1._wp, 2._wp, 3._wp]
    wavevector = [.1_wp, .2_wp, .3_wp]
    amplitude = [1._wp, 2._wp, 3._wp]

    res1 = plane_wave(&
         r          , &
         wavenumber1, &
         amplitude  , &
         wavevector)
    res2 = plane_wave(&
         r          , &
         wavenumber2, &
         amplitude  , &
         wavevector)
    res3 = plane_wave(      &
         r                , &
         wavenumber1      , &
         amplitude        , &
         wavevector       , &
         angular_frequency, &
         time)
    
    if (.not. all(is_close(&
         res1%re, [1._wp, 2._wp, 3._wp]*0.9902159962126371))   &
         .or. .not. all(is_close(&
         res1%im, [1._wp, 2._wp, 3._wp]*0.1395431146442365))   &
         .or. .not. all(is_close(&
         res2%re, [1._wp, 2._wp, 3._wp]*0.8608524311310889))   &
         .or. .not. all(is_close(&
         res2%im, [1._wp, 2._wp, 3._wp]*0.1213129559091667))   &
         .or. .not. all(is_close(&
         res3%re, [1._wp, 2._wp, 3._wp]*0.9872272833756269))   &
         .or. .not. all(is_close(&
         res3%im, [1._wp, 2._wp, 3._wp]*(-0.159318206614246))) )  then
       msg = 'Incorrect test results.'
       print *, res1
       print *, res2
       print *, res3
       call print_error_message(module_name, test_name, msg, degree=2)
    end if
    
    call print_status_done(test_nr)
  end subroutine test_plane_wave
  

  !!----------------------------------------------------------------------------
end program main
