!!============================================================================
! This program runs tests for the module 'mesh_mod'.
!
!   First a series of subroutines are defines for testing specific parts of 
!   the module. Then a sequence of these routines are called in the main
!   program. The assertions are defined in the respective subroutines.
!
! Last edited: March 7th 2021.
!!==============================================================================


!!===========!!
! Subroutines !
!!============!=================================================================
subroutine test_initialise(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start
  use test_utilities   , only: print_status_done
  use test_utilities   , only: check_mesh_calculations
  use mesh_mod         , only: mesh_type

  implicit none
  integer, intent(in)           :: num_tests
  integer, intent(inout)        :: test_nr
  ! Variables for internal use -------------------------------------------------
  type (mesh_type)              :: mesh
  character(len=:), allocatable :: filename
  character(len=:), allocatable :: file_format
  logical                       :: closed_surface
  integer                       :: edge_order
  integer                       :: spatial_dim
  integer                       :: num_face_vertices
  integer                       :: num_handles
  integer                       :: num_apertures
  integer                       :: num_boundary_edges
  integer                       :: correct_num_faces
  integer                       :: correct_num_edges
  integer                       :: correct_num_vertices
  integer                       :: correct_num_apertures
  integer                       :: correct_num_boundary_edges
  integer                       :: correct_num_handles
  real(wp)                      :: correct_face_area
  real(wp)                      :: correct_surface_area
  real(wp)                      :: correct_volume
  integer :: i

  call print_status_start(num_tests, test_nr , 'mesh_mod', &
       'initialise')

  !!----------------!!
  ! Test tetrahedron !
  !------------------!----------------------------------------------------------
  filename = 'meshes/tetrahedron.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_face_vertices = 3
  
  call mesh%initialise(&
       filename      , &
       file_format   , &
       closed_surface, &
       edge_order    , &
       spatial_dim   , &
       num_face_vertices)

  correct_num_faces = 4
  correct_num_edges = 6
  correct_num_vertices = 4
  correct_num_handles = 0
  correct_num_apertures = 0
  correct_num_boundary_edges = 0
  ! Surface area of a regular tetrahedron with sides a = 2*sqrt(2)
  correct_face_area = 2._wp*sqrt(3._wp)
  correct_surface_area = 8._wp*sqrt(3._wp) 
  ! Volume of a regular tetrahedron with sides
  correct_volume = 8._wp/3._wp 

  call check_mesh_calculations(&
       mesh                      , &
       correct_num_faces         , &
       correct_num_edges         , &
       correct_num_vertices      , &
       correct_num_handles       , &
       correct_num_apertures     , &
       correct_num_boundary_edges, &
       correct_face_area         , &
       correct_surface_area      , &
       correct_volume=correct_volume)
  
  !!----------------!!
  ! Test closed cube !
  !------------------!----------------------------------------------------------
  filename = 'meshes/cube_closed.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_face_vertices = 3
  
  call mesh%deallocate_attributes()
  call mesh%initialise(&
       filename      , &
       file_format   , &
       closed_surface, &
       edge_order    , &
       spatial_dim   , &
       num_face_vertices)

  correct_num_faces = 12
  correct_num_edges = 18
  correct_num_vertices = 8
  correct_num_handles = 0
  correct_num_apertures = 0
  correct_num_boundary_edges = 0
  correct_face_area = 0.5_wp
  ! Surface area of a cube with sides 1.0
  correct_surface_area = 6._wp
  ! Volume of a cube with sides 1.0
  correct_volume = 1._wp 

  call check_mesh_calculations(&
       mesh                      , &
       correct_num_faces         , &
       correct_num_edges         , &
       correct_num_vertices      , &
       correct_num_handles       , &
       correct_num_apertures     , &
       correct_num_boundary_edges, &
       correct_face_area         , &
       correct_surface_area      , &
       correct_volume=correct_volume)

  !!--------------!!
  ! Test open cube !
  !----------------!------------------------------------------------------------
  filename = 'meshes/cube_open.msh'
  file_format = 'gmsh2'
  closed_surface = .false.
  edge_order = 0
  spatial_dim = 3
  num_face_vertices = 3
  num_handles = 0
  num_apertures = 1
  num_boundary_edges = 3
  
  call mesh%deallocate_attributes()
  call mesh%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)

  correct_num_faces = 11
  correct_num_edges = 18
  correct_num_vertices = 8
  correct_num_handles = 0
  correct_num_apertures = 1
  correct_num_boundary_edges = 3
  correct_face_area = 0.5_wp
  ! Surface area of a cube with sides 1.0
  correct_surface_area = 5.5_wp

  call check_mesh_calculations(&
       mesh                      , &
       correct_num_faces         , &
       correct_num_edges         , &
       correct_num_vertices      , &
       correct_num_handles       , &
       correct_num_apertures     , &
       correct_num_boundary_edges, &
       correct_face_area         , &
       correct_surface_area)

  call print_status_done(test_nr)
end subroutine test_initialise

!!------------------------------------------------------------------------------

subroutine test_face_area(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start, print_status_done
  use mesh_mod         , only: mesh_type
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  type (mesh_type)       :: mesh

  call print_status_start(num_tests, test_nr , 'mesh_mod', 'face_area')
  
    call print_status_done(test_nr)
end subroutine test_face_area

!!------------------------------------------------------------------------------

subroutine test_surface_area(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start, print_status_done
  use mesh_mod         , only: mesh_type
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  type (mesh_type)       :: mesh
  real(wp)               :: surface_area
  real(wp)               :: correct_result

  call print_status_start(num_tests, test_nr , 'mesh_mod', 'surface_area')
  
  correct_result = 8._wp*sqrt(3._wp) ! Surface area of a regular tetrahedron 
  ! with sides a = 2*sqrt(2)
  call mesh%initialise_tetrahedron()
  surface_area = mesh%surface_area()
  if (surface_area /= correct_result) then
     print *, ''
     print *, 'Error: test_surface_area in mesh_test.f90 failed..'
     print *, '  Estimated surface area: ', surface_area, 'Correct result: ', &
          correct_result
     stop 1
  end if
  
  call print_status_done(test_nr)
end subroutine test_surface_area

!!------------------------------------------------------------------------------

subroutine test_volume(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start, print_status_done
  use mesh_mod         , only: mesh_type
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  type (mesh_type)       :: mesh
  real(wp)               :: volume
  real(wp)               :: correct_result

  call print_status_start(num_tests, test_nr , 'mesh_mod', 'volume')
  
  correct_result = 8._wp/3._wp ! Volume of a regular tetrahedron with sides
  ! a = 2*sqrt(2)
  call mesh%initialise_tetrahedron()
  volume = mesh%volume()
  if (volume /= correct_result) then
     print *, ''
     print *, 'Error: test_volume in mesh_test.f90 failed..'
     print *, '  Estimated volume: ', volume, 'Correct result: ', &
          correct_result
     stop 1
  end if

  call print_status_done(test_nr)
end subroutine test_volume

!!------------------------------------------------------------------------------

subroutine test_initialise_tetrahedron(num_tests, test_nr)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start 
  use test_utilities   , only: print_status_done
  use mesh_mod         , only: mesh_type
  implicit none
  integer, intent(in)    :: num_tests
  integer, intent(inout) :: test_nr
  type (mesh_type)       :: mesh
  integer                :: i, num_faces, num_handles
  real(wp)               :: area, surface_area, volume
  real(wp), dimension(3) :: norm, unit_norm
  integer                :: correct_num_handles
  integer                :: correct_num_faces
  real(wp)               :: correct_face_area

  call print_status_start(num_tests, test_nr , 'mesh_mod', &
       'initialise_tetrahedron')
  
  correct_num_handles = 0
  correct_num_faces = 4
  correct_face_area = 2._wp*sqrt(3._wp)
  call mesh%initialise_tetrahedron()
  num_handles = mesh%get_num_handles()
  num_faces = mesh%get_num_faces()
  do i = 1, num_faces
     area = mesh%face_area(i)
     if (area /= correct_face_area) then
     print *, ''
        print *, 'Error: test_initialise_tetrahedron in mesh_test.f90', &
        ' failed..'
        print *, '  Estimated face area: ', area, 'Correct result: ', &
             correct_face_area
        stop 1
     end if
  end do
  if (num_handles /= correct_num_handles) then
     print *, ''
     print *, 'Error: test_initialise_tetrahedron in mesh_test.f90', &
     ' failed..'
     print *, '  Estimated number of handles: ', num_handles, &
          'Correct result: ', correct_num_handles
     stop 1
  end if
  if (num_faces /= correct_num_faces) then
     print *, ''
     print *, 'Error: test_initialise_tetrahedron in mesh_test.f90', &
     ' failed..'
     print *, '  Estimated number of faces: ', num_faces, &
          'Correct result: ', correct_num_faces
     stop 1
  end if

  call print_status_done(test_nr)
end subroutine test_initialise_tetrahedron

!!------------------------------------------------------------------------------

subroutine test_write_mesh(num_tests, test_nr, &
     tetrahedron, &
     closed_cube, &
     open_cube  , &
     sphere)
  use working_precision, only: wp
  use test_utilities   , only: print_status_start 
  use test_utilities   , only: print_status_done
  use mesh_mod         , only: mesh_type

  implicit none
  integer         , intent(in)    :: num_tests
  integer         , intent(inout) :: test_nr
  type (mesh_type), intent(in)    :: tetrahedron
  type (mesh_type), intent(in)    :: closed_cube
  type (mesh_type), intent(in)    :: open_cube
  type (mesh_type), intent(in)    :: sphere
  ! Variables for internal use -------------------------------------------------
  character(len=255) :: filename

  call print_status_start(num_tests, test_nr , 'mesh_mod', &
       'write_mesh')

  filename = 'tetrahedron_processed.txt'
  call tetrahedron%write_mesh(filename)
  filename = 'closed_cube_processed.txt'
  call closed_cube%write_mesh(filename)
  filename = 'open_cube_processed.txt'
  call open_cube%write_mesh(filename)
  filename = 'sphere_processed.txt'
  call sphere%write_mesh(filename)
  
  call print_status_done(test_nr)
end subroutine test_write_mesh
!!------------------------------------------------------------------------------


!!=============!
! Main program !
!!=============!================================================================
program main
  use working_precision, only: wp
  use mesh_mod, only: mesh_type
  use test_utilities   , only: print_status_start 
  use test_utilities   , only: print_status_done

  implicit none
  integer, parameter :: num_tests = 5
  integer            :: test_nr = 1
  ! Variables for making mesh objects ------------------------------------------
  type (mesh_type)       :: tetrahedron
  type (mesh_type)       :: closed_cube
  type (mesh_type)       :: open_cube
  type (mesh_type)       :: sphere
  integer                :: edge_order
  integer                :: spatial_dim
  integer                :: num_face_vertices
  integer                :: num_handles
  integer                :: num_apertures
  integer                :: num_boundary_edges
  logical                :: closed_surface
  character(len=255)     :: filename
  character(len=255)     :: file_format
  

  filename = 'meshes/tetrahedron.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  spatial_dim = 3
  num_face_vertices = 3
  call tetrahedron%initialise(&
       filename      , &
       file_format   , &
       closed_surface, &
       edge_order    , &
       spatial_dim   , &
       num_face_vertices)

  filename = 'meshes/cube_closed.msh'
  call closed_cube%initialise(&
       filename      , &
       file_format   , &
       closed_surface, &
       edge_order    , &
       spatial_dim   , &
       num_face_vertices)

  filename = 'meshes/gmsh_sphere_1ord.msh'
  call sphere%initialise(&
       filename      , &
       file_format   , &
       closed_surface, &
       edge_order    , &
       spatial_dim   , &
       num_face_vertices)
  
  filename = 'meshes/cube_open.msh'
  closed_surface = .false.
  edge_order = 0
  spatial_dim = 3
  num_face_vertices = 3
  num_handles = 0
  num_apertures = 1
  num_boundary_edges = 3
  call open_cube%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)


  call test_initialise(num_tests, test_nr)
  call test_initialise_tetrahedron(num_tests, test_nr)
  call test_volume(num_tests, test_nr)
  call test_surface_area(num_tests, test_nr)
  call test_write_mesh(num_tests, test_nr, &
       tetrahedron, &
       closed_cube, &
       open_cube  , &
       sphere)
  
contains

end program main
  

