!!==============================================================================
! This program runs convergence tests of various Gaussian quadratures performed
! by PMCHW_RWG_mod.f90. 
!
! Last edited: January 30th 2021.
!!==============================================================================


!!===========!!
! Subroutines !
!!============!=================================================================

!!============!!
! Main program !
!!=============!================================================================
program main
  ! Working precision
  use working_precision      , only: wp
  use iso_fortran_env        , only: real64
  ! Derived types
  use mesh_mod               , only: mesh_type
  use RWG_basis_mod          , only: RWG_basis_type
  use PMCHW_RWG_mod          , only: PMCHW_RWG_type
  ! Constants
  use constants_mod          , only: PI
  use constants_mod          , only: ZERO_CMPLX
  use constants_mod          , only: PERMEABILITY_VACUUM
  use constants_mod          , only: PERMITIVITY_VACUUM
  ! Gaussian quadrature formulas
  use gauss_quad_formulas_mod, only: GQF_triangle_3pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_4pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_6pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_7pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_9pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_12pnt
  use gauss_quad_formulas_mod, only: GQF_triangle_13pnt
  ! Routines for performing the quadratures
  use PMCHW_RWG_mod          , only: eval_green_func_integrals
  use PMCHW_RWG_mod          , only: eval_outer_integrals
  use PMCHW_RWG_mod          , only: Cauchy_principal_value
  use PMCHW_RWG_mod          , only: face_pair_integral_EFIE
  use PMCHW_RWG_mod          , only: face_pair_integral_MFIE
  use PMCHW_RWG_mod          , only: calc_edge_unit_normals
  use PMCHW_RWG_mod          , only: eval_subtracted_terms
  ! Writing routines
  use io_mod                 , only: r8mat_write
  !-----------------------------------------------------------------------------
  
  implicit none
  ! Switch to use direct base-by-base integration or faster face-by-face.
  logical, parameter        :: direct = .false.
  character(*), parameter   :: case_name &
       = '../results/conv'
  ! Variables for generating mesh and basis ------------------------------------
  type (mesh_type)                       :: mesh
  type (RWG_basis_type)                  :: closed_cube
  character(len=:), allocatable          :: filename
  character(len=:), allocatable          :: file_format
  logical                                :: closed_surface
  integer                                :: edge_order
  integer                                :: num_handles
  integer                                :: num_apertures
  integer                                :: num_boundary_edges
  ! Variables for generating scattering problem example ------------------------
  type (PMCHW_RWG_type)             :: PMCHW_closed_cube
  complex(wp), dimension(2)         :: permeabilities
  complex(wp), dimension(2)         :: permitivities
  complex(wp), dimension(3, 2)      :: E_inc_ampl
  complex(wp), dimension(3, 2)      :: H_inc_ampl
  real(wp)   , dimension(3, 2)      :: E_inc_direction
  real(wp)   , dimension(3, 2)      :: H_inc_direction
  real(wp)                          :: angular_frequency
  real(wp)                          :: wavelength
  integer    , dimension(2)         :: inc_field_type
  ! Useful constants -----------------------------------------------------------
  integer, parameter :: NUM_FACE_VERTICES = 3
  integer, parameter :: SPATIAL_DIM       = 3

  !!-----------------------------!!
  ! Scattering example parameters !
  !-------------------------------!---------------------------------------------
  ! Permeability and permititvity of region 1 (outer region)
  permeabilities(1) = cmplx(1._wp, 0._wp)*PERMEABILITY_VACUUM
  permitivities(1) = cmplx(1._wp, 0._wp)*PERMITIVITY_VACUUM
  ! Region 2, inner region. Assuming non-magnetic material, such that
  ! the relative permeability equals 1. In this case we have a complex
  ! permitivity, depending on the angular frequency of the EM-wave. 
  ! Below: High permitivity scatterer (refractive index = 4)
  permeabilities(2) = cmplx(1._wp, 0._wp)*PERMEABILITY_VACUUM
  permitivities(2) = cmplx(16._wp, 0._wp)*PERMITIVITY_VACUUM
  wavelength = 2.e-0_wp ! Wavelength in meter in vacuum
  ! With radius of the sphere, a, equal to 1 m this makes the relationship
  ! a/wavelength = 1/2
  angular_frequency = 2._wp*PI*3.e8_wp/wavelength

  ! Incoming plane waves traveling in the positive z-direction
  inc_field_type(1) = 1
  E_inc_direction(:, 1) = [ 0._wp, 0._wp, 1._wp ]
  H_inc_direction(:, 1) = [ 0._wp, 0._wp, 1._wp ]
  ! Incoming plane waves traveling in the positive z-direction
  inc_field_type(2) = 1
  E_inc_direction(:, 2) = [ 1._wp, 0._wp, 0._wp ]
  H_inc_direction(:, 2) = [ 1._wp, 0._wp, 0._wp ]

  ! Polarization and intensity.
  ! Illumination intensity is equal to unity.
  ! E-field is x-polarized (H-field y-polarized)
  E_inc_ampl(:, 1) = [      &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp)         , &
       cmplx(0._wp, 0._wp) ]
  H_inc_ampl(:, 1) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp) ]
  ! E-field is y-polarized (H-field z-polarized)
  E_inc_ampl(:, 2) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp), &
       cmplx(0._wp, 0._wp) ]
  H_inc_ampl(:, 2) = [      &
       cmplx(0._wp, 0._wp)         , &
       cmplx(0._wp, 0._wp)         , &
       cmplx(1._wp, 0._wp) ]
  
  !!---------------------!!
  ! Construct closed cube !
  !-----------------------!-----------------------------------------------------
  filename = 'meshes/cube_closed.msh'
  file_format = 'gmsh2'
  closed_surface = .true.
  edge_order = 0
  num_handles = 0
  num_apertures = 0
  num_boundary_edges = 0
  call mesh%initialise(&
       filename                      , &
       file_format                   , &
       closed_surface                , &
       edge_order                    , &
       spatial_dim                   , &
       num_face_vertices             , &
       num_handles_in=num_handles    , &
       num_apertures_in=num_apertures, &
       num_boundary_edges_in=num_boundary_edges)
  call closed_cube%initialise(mesh)
  call PMCHW_closed_cube%initialise(&
       closed_cube   , &
       permeabilities, &
       permitivities , &
       angular_frequency)

  call convergence_test(&
       PMCHW_closed_cube, &
       direct           , &
       E_inc_ampl       , &
       H_inc_ampl       , &
       E_inc_direction  , &
       H_inc_direction  , &
       case_name)

  !=======!=========================!==========================================!
contains  ! /\/\/\/\/\/\/\/\/\/\/\/\!/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\!
  !=======!=========================!==========================================!

  subroutine convergence_test(&
       PMCHW          , &
       direct         , &
       E_inc_ampl     , &
       H_inc_ampl     , &
       E_inc_direction, &
       H_inc_direction, &
       case_name)
    type (PMCHW_RWG_type)              , intent(in) :: PMCHW
    logical                            , intent(in) :: direct
    complex(wp), dimension(SPATIAL_DIM), intent(in) :: E_inc_ampl
    complex(wp), dimension(SPATIAL_DIM), intent(in) :: H_inc_ampl
    real(wp)   , dimension(SPATIAL_DIM), intent(in) :: E_inc_direction
    real(wp)   , dimension(SPATIAL_DIM), intent(in) :: H_inc_direction
    character(*)                       , intent(in) :: case_name
    ! Variables for internal use -----------------------------------------------
    character(len=255)                        :: filename1
    character(len=255)                        :: filename2
    integer, parameter                        :: GQF_XI_IDX = 2
    integer, parameter                        :: GQF_ETA_IDX = 3
    integer, parameter                        :: GQF_ZETA_IDX = 4
    integer, parameter                        :: NUM_QUADRATURES = 6
    integer, parameter                        :: NUM_EFIE_QUADS = 9
    integer, parameter                        :: NUM_MFIE_QUADS = 18
    complex(wp), dimension(:, :), allocatable :: quadratures
    complex(wp), dimension(:, :), allocatable :: EFIE_quads
    complex(wp), dimension(:, :), allocatable :: MFIE_quads
    complex(wp), dimension(:)   , allocatable :: D_mn_pq
    complex(wp), dimension(:)   , allocatable :: K_mn_pq
    complex(wp)                               :: subtr_terms_D_mn_pq
    complex(wp)                               :: subtr_terms_K_mn_pq
    complex(wp)                               :: wavenumber
    real(wp)   , dimension(:, :), allocatable :: table
    real(wp)   , dimension(:, :), allocatable :: table2
    real(wp)   , dimension(:, :), allocatable :: quad_pnts_src
    real(wp)   , dimension(:, :), allocatable :: quad_pnts_obs
    real(wp)   , dimension(:, :), allocatable :: gauss_quad_formula
    real(wp)   , dimension(NUM_FACE_VERTICES, SPATIAL_DIM) :: face_coords
    real(wp)   , dimension(NUM_FACE_VERTICES, SPATIAL_DIM) ::edge_unit_normals_p
    real(wp)   , dimension(NUM_FACE_VERTICES, SPATIAL_DIM) ::edge_unit_normals_q
    real(wp)   , dimension(NUM_FACE_VERTICES) :: edge_lengths_p
    real(wp)   , dimension(NUM_FACE_VERTICES) :: edge_lengths_q
    real(wp)   , dimension(SPATIAL_DIM)       :: face_unit_normal_p
    real(wp)   , dimension(SPATIAL_DIM)       :: face_unit_normal_q
    real(wp)   , dimension(SPATIAL_DIM)       :: r1_obs
    real(wp)   , dimension(SPATIAL_DIM)       :: r2_obs
    real(wp)   , dimension(SPATIAL_DIM)       :: r3_obs
    real(wp)   , dimension(SPATIAL_DIM)       :: r1_src
    real(wp)   , dimension(SPATIAL_DIM)       :: r2_src
    real(wp)   , dimension(SPATIAL_DIM)       :: r3_src
    real(wp)   , dimension(SPATIAL_DIM)       :: p_m
    real(wp)   , dimension(SPATIAL_DIM)       :: p_n
    real(wp)   , dimension(SPATIAL_DIM)       :: obs_pnt
    real(wp)                                  :: xi
    real(wp)                                  :: eta
    real(wp)                                  :: zeta
    real(wp)                                  :: face_area_p
    real(wp)                                  :: face_area_q
    integer                                   :: region
    integer                                   :: obs_face
    integer                                   :: src_face
    integer                                   :: p_m_idx
    integer                                   :: p_n_idx
    integer                                   :: num_quad_formulas
    integer                                   :: num_quad_pnts
    integer                                   :: i_0
    integer                                   :: i
    integer                                   :: j
    logical                                   :: CAUCHY
    logical                                   :: faces_are_close
    
    ! User options -------------------------------------------------------------
    CAUCHY = .false.
    faces_are_close = .true.
    region = 1
    obs_face = 1 ! 1 or 6
    src_face = 10 ! 6. 10 for "far" away
    p_m_idx = 3 ! 3 or 1
    p_n_idx = 2 ! 2. 3 for "far" away 
    i_0 = 1               ! < 8 and > 0
    num_quad_formulas = 7 ! < 8 and > 0 and >= i_0
    !---------------------------------------------------------------------------
    
    wavenumber = PMCHW%angular_frequency*sqrt(&
         PMCHW%permeabilities(region)*PMCHW%permitivities(region))

    face_coords = PMCHW%RWG_basis%mesh%get_face_coords(obs_face)
    r1_obs = face_coords(1, :)
    r2_obs = face_coords(2, :)
    r3_obs = face_coords(3, :)
    p_m = face_coords(p_m_idx, :)
    face_coords = PMCHW%RWG_basis%mesh%get_face_coords(src_face)
    r1_src = face_coords(1, :)
    r2_src = face_coords(2, :)
    r3_src = face_coords(3, :)
    p_n = face_coords(p_n_idx, :)

    ! Centroid of face
    obs_pnt = 1._wp/3._wp*(r1_obs + r2_obs + r3_obs)

    allocate(quadratures(NUM_QUADRATURES, num_quad_formulas))
    allocate(EFIE_quads(NUM_EFIE_QUADS, num_quad_formulas))
    allocate(MFIE_quads(NUM_MFIE_QUADS, num_quad_formulas))
    allocate(D_mn_pq(num_quad_formulas))
    allocate(K_mn_pq(num_quad_formulas))
    do i = i_0, num_quad_formulas
       if (allocated(gauss_quad_formula)) deallocate(gauss_quad_formula)
       select case (i)
       case (1)
          allocate(gauss_quad_formula, source=GQF_triangle_3pnt)
       case (2)
          allocate(gauss_quad_formula, source=GQF_triangle_4pnt)
       case (3)
          allocate(gauss_quad_formula, source=GQF_triangle_6pnt)
       case (4)
          allocate(gauss_quad_formula, source=GQF_triangle_7pnt)
       case (5)
          allocate(gauss_quad_formula, source=GQF_triangle_9pnt)
       case (6)
          allocate(gauss_quad_formula, source=GQF_triangle_12pnt)
       case (7)
          allocate(gauss_quad_formula, source=GQF_triangle_13pnt)
       end select  

       num_quad_pnts = size(gauss_quad_formula, dim=1)
       if (allocated(quad_pnts_src)) deallocate(quad_pnts_src)
       if (allocated(quad_pnts_obs)) deallocate(quad_pnts_obs)
       allocate(quad_pnts_src(num_quad_pnts, SPATIAL_DIM))
       allocate(quad_pnts_obs(num_quad_pnts, SPATIAL_DIM))
       ! Calculate quadrature points
       do j = 1, num_quad_pnts
          xi = gauss_quad_formula(j, GQF_XI_IDX)
          eta = gauss_quad_formula(j, GQF_ETA_IDX)
          zeta = gauss_quad_formula(j, GQF_ZETA_IDX)
          quad_pnts_src(j, :) = xi*r1_src + eta*r2_src + zeta*r3_src
          quad_pnts_obs(j, :) = xi*r1_obs + eta*r2_obs + zeta*r3_obs
       end do ! j
       
       ! Evaluating inner integrals only
       call eval_green_func_integrals(&
            quadratures(1, i) , &
            quadratures(2, i) , &
            quadratures(3, i) , &
            quadratures(4, i) , &
            quadratures(5, i) , &
            quadratures(6, i) , &
            obs_pnt           , &
            quad_pnts_src     , &
            gauss_quad_formula, &
            wavenumber        , &
            faces_are_close   , &
            faces_are_in_same_plane=(obs_face == src_face))

       ! Evaluating double integrals
       call eval_outer_integrals(&
            EFIE_quads(:, i)  , &
            MFIE_quads(:, i)  , &
            quad_pnts_src     , &
            quad_pnts_obs     , &
            gauss_quad_formula, &
            faces_are_close   , &
            obs_face          , &
            src_face          , &
            wavenumber)
       D_mn_pq(i) = face_pair_integral_EFIE(&
            r1_obs    , &
            r2_obs    , &
            r3_obs    , &
            r1_src    , &
            r2_src    , &
            r3_src    , &
            p_m       , &
            p_n       , &
            wavenumber, &
            EFIE_QUADS(:, i))
       K_mn_pq(i) = face_pair_integral_MFIE(&
            r1_obs, &
            r2_obs, &
            r3_obs, &
            r1_src, &
            r2_src, &
            r3_src, &
            p_m   , &
            p_n   , &
            MFIE_QUADS(:, i))
       
       ! If faces are close, add integration over terms subtracted from the
       ! Green's function to regularise it.
       if (faces_are_close) then
          face_area_q = PMCHW%RWG_basis%mesh%face_area(src_face)
          face_unit_normal_q = PMCHW%RWG_basis%mesh%face_unit_normal(src_face)
          edge_lengths_q = PMCHW%get_edge_lengths(src_face)
          edge_unit_normals_q = calc_edge_unit_normals(&
               edge_lengths_q    , &
               face_unit_normal_q, &
               r1_src            , &
               r2_src            , &
               r3_src)
          face_area_p = PMCHW%RWG_basis%mesh%face_area(obs_face)
          face_unit_normal_p = PMCHW%RWG_basis%mesh%face_unit_normal(obs_face)
          edge_lengths_p = PMCHW%get_edge_lengths(obs_face)
          edge_unit_normals_p = calc_edge_unit_normals(&
               edge_lengths_p    , &
               face_unit_normal_p, &
               r1_obs            , &
               r2_obs            , &
               r3_obs)
!!$             print *, '-----------------------------'
!!$             print *, ' FACE:', obs_face
!!$             print *, face_area_p
!!$             print *, face_unit_normal_p
!!$             print *, edge_lengths_p
!!$             print *, edge_unit_normals_p(1, :)
!!$             print *, edge_unit_normals_p(2, :)
!!$             print *, edge_unit_normals_p(3, :)
!!$             print *, ' FACE:', src_face
!!$             print *, face_area_q
!!$             print *, face_unit_normal_q
!!$             print *, edge_lengths_q
!!$             print *, edge_unit_normals_q(1, :)
!!$             print *, edge_unit_normals_q(2, :)
!!$             print *, edge_unit_normals_q(3, :)
!!$          
          call eval_subtracted_terms(&
               subtr_terms_D_mn_pq, &
               subtr_terms_K_mn_pq, &
               wavenumber         , &
               r1_obs             , &
               r2_obs             , &
               r3_obs             , &
               r1_src             , &
               r2_src             , &
               r3_src             , &
               p_m                , &
               p_n                , &
               gauss_quad_formula , &
               obs_face           , &
               src_face           , &
               face_area_p        , &
               face_area_q        , &
               face_unit_normal_p , &
               face_unit_normal_q , &
               edge_lengths_p     , &
               edge_lengths_q     , &
               edge_unit_normals_p, &
               edge_unit_normals_q)
!!$          print *, 'i D K', i, D_mn_pq(i), K_mn_pq(i)
!!$          print *, 'i SD SK', i, subtr_terms_D_mn_pq, subtr_terms_K_mn_pq
          D_mn_pq(i) = D_mn_pq(i) + subtr_terms_D_mn_pq
          K_mn_pq(i) = K_mn_pq(i) + subtr_terms_K_mn_pq
       end if
    end do ! i
    

    ! Alternativ addition:
    ! Top row with information on wavelength, object, permitivity, etc.
    !table(1, 1) = 
    allocate(table(num_quad_formulas, NUM_QUADRATURES*2))
    do i = 1, NUM_QUADRATURES
       table(:, 1 + 2*(i - 1)) = quadratures(i, :)%re
       table(:, 2 + 2*(i - 1)) = quadratures(i, :)%im
    end do
    allocate(table2(num_quad_formulas, 4))
    table2(:, 1) = D_mn_pq%re
    table2(:, 2) = D_mn_pq%im
    table2(:, 3) = K_mn_pq%re
    table2(:, 4) = K_mn_pq%im

    if (wp == real64) then
       write(filename1, '(a,a)') case_name, '_inner_intgr.txt'
       write(filename2, '(a,a)') case_name, '_double_intgr.txt'
       call r8mat_write(filename1, num_quad_formulas, NUM_QUADRATURES*2, &
            table)
       call r8mat_write(filename2, num_quad_formulas, 4, &
            table2)
    else
       print *, 'No write routine for current precision.'
    end if
  
!!$    
!!$    if (direct) then
!!$       call PMCHW_closed_cube%calc_q_vectors_direct(&
!!$            GQF_triangle_3pnt, &
!!$            E_inc_ampl       , &
!!$            H_inc_ampl       , &
!!$            E_inc_direction  , &
!!$            H_inc_direction  , &
!!$            inc_field_type)
!!$    else
!!$       call PMCHW_closed_cube%calc_q_vectors(&
!!$            GQF_triangle_3pnt, &
!!$            E_inc_ampl       , &
!!$            H_inc_ampl       , &
!!$            E_inc_direction  , &
!!$            H_inc_direction  , &
!!$            inc_field_type)
!!$    end if
!!$    call PMCHW_closed_cube%calc_PMCHW_matrix(GQF_triangle_3pnt, direct)
!!$    call PMCHW_closed_cube%solve_matrix_equation()
  end subroutine convergence_test
  !!----------------------------------------------------------------------------

end program main
